import torch, itertools, functools
import torch.nn as nn

from torch.utils.checkpoint import checkpoint

from .utils import cut_to_match, size_is_pow2
from .ops import TrivialUpsample, TrivialDownsample, NoOp, UGroupNorm, u_group_norm
from .blocks import UnetDownBlock, UnetUpBlock, ThinUnetDownBlock, ThinUnetUpBlock

fat_setup = {
    'gate': nn.PReLU,
    'norm': nn.InstanceNorm2d,
    'upsample': TrivialUpsample,
    'downsample': TrivialDownsample,
    'down_block': UnetDownBlock,
    'up_block': UnetUpBlock,
    'dropout': NoOp,
    'padding': False,
    'bias': True
}

thin_setup = {
    'gate': nn.PReLU,
    'norm': nn.InstanceNorm2d,
    'upsample': TrivialUpsample,
    'downsample': TrivialDownsample,
    'down_block': ThinUnetDownBlock,
    'up_block': ThinUnetUpBlock,
    'dropout': NoOp,
    'padding': False,
    'bias': True
}

def checkpointed(cls):
    assert issubclass(cls, torch.nn.Module)

    #@functools.wraps(cls)
    class Checkpointed(cls):
        def forward(self, *args, **kwargs):
            super_fwd = super(Checkpointed, self).forward
            if any((torch.is_tensor(arg) and arg.requires_grad) for arg in args):
                return checkpoint(super_fwd, *args, **kwargs)
            else:
                return super_fwd(*args, **kwargs)

    return Checkpointed

class Unet(nn.Module):
    def __init__(self, in_features=1, up=None, down=None,
                 size=5, setup=fat_setup):

        super(Unet, self).__init__()

        if not len(down) == len(up) + 1:
            raise ValueError("`down` must be 1 item longer than `up`")

        self.up = up
        self.down = down
        self.in_features = in_features

        DownBlock = setup['down_block']
        UpBlock = setup['up_block']

        if 'checkpointed' in setup and setup['checkpointed']:
            UpBlock = checkpointed(UpBlock)
            DownBlock = checkpointed(DownBlock)

        down_dims = [in_features] + down
        self.path_down = nn.ModuleList()
        for i, (d_in, d_out) in enumerate(zip(down_dims[:-1], down_dims[1:])):
            block = DownBlock(
                d_in, d_out, size=size, name=f'down_{i}', setup=setup, is_first=i==0,
            )
            self.path_down.append(block)

        bot_dims = [down[-1]] + up
        hor_dims = down_dims[-2::-1]
        self.path_up = nn.ModuleList()
        for i, (d_bot, d_hor, d_out) in enumerate(zip(bot_dims, hor_dims, up)):
            block = UpBlock(
                d_bot, d_hor, d_out, size=size, name=f'up_{i}', setup=setup
            )
            self.path_up.append(block)

        self.n_params = 0
        for param in self.parameters():
            self.n_params += param.numel()


    # @dimchecked
    def forward(self, inp):
        if inp.size(1) != self.in_features:
            fmt = "Expected {} feature channels in input, got {}"
            msg = fmt.format(self.in_features, inp.size(1))
            raise ValueError(msg)

        features = [inp]
        for i, layer in enumerate(self.path_down):
            features.append(layer(features[-1]))

        f_bot = features[-1]
        features_horizontal = features[-2::-1]

        for layer, f_hor in zip(self.path_up, features_horizontal):
            f_bot = layer(f_bot, f_hor)

        return f_bot



from typing import Callable, Optional

import torch
import torch.nn.functional as F
import torchvision
from torch import nn
from torch.nn.modules.utils import _pair
from torchvision.models import resnet

from omegaconf import OmegaConf

# coordinates system
#  ------------------------------>  [ x: range=-1.0~1.0; w: range=0~W ]
#  | -----------------------------
#  | |                           |
#  | |                           |
#  | |                           |
#  | |         image             |
#  | |                           |
#  | |                           |
#  | |                           |
#  | |---------------------------|
#  v
# [ y: range=-1.0~1.0; h: range=0~H ]


def get_patches(
    tensor: torch.Tensor, required_corners: torch.Tensor, ps: int
) -> torch.Tensor:
    c, h, w = tensor.shape
    corner = (required_corners - ps / 2 + 1).long()
    corner[:, 0] = corner[:, 0].clamp(min=0, max=w - 1 - ps)
    corner[:, 1] = corner[:, 1].clamp(min=0, max=h - 1 - ps)
    offset = torch.arange(0, ps)

    kw = {"indexing": "ij"} if torch.__version__ >= "1.10" else {}
    x, y = torch.meshgrid(offset, offset, **kw)
    patches = torch.stack((x, y)).permute(2, 1, 0).unsqueeze(2)
    patches = patches.to(corner) + corner[None, None]
    pts = patches.reshape(-1, 2)
    sampled = tensor.permute(1, 2, 0)[tuple(pts.T)[::-1]]
    sampled = sampled.reshape(ps, ps, -1, c)
    assert sampled.shape[:3] == patches.shape[:3]
    return sampled.permute(2, 3, 0, 1)


def simple_nms(scores: torch.Tensor, nms_radius: int):
    """Fast Non-maximum suppression to remove nearby points"""

    zeros = torch.zeros_like(scores)
    max_mask = scores == torch.nn.functional.max_pool2d(
        scores, kernel_size=nms_radius * 2 + 1, stride=1, padding=nms_radius
    )

    for _ in range(2):
        supp_mask = (
            torch.nn.functional.max_pool2d(
                max_mask.float(),
                kernel_size=nms_radius * 2 + 1,
                stride=1,
                padding=nms_radius,
            )
            > 0
        )
        supp_scores = torch.where(supp_mask, zeros, scores)
        new_max_mask = supp_scores == torch.nn.functional.max_pool2d(
            supp_scores, kernel_size=nms_radius * 2 + 1, stride=1, padding=nms_radius
        )
        max_mask = max_mask | (new_max_mask & (~supp_mask))
    return torch.where(max_mask, scores, zeros)


class DKD(nn.Module):
    def __init__(
        self,
        radius: int = 2,
        top_k: int = 0,
        scores_th: float = 0.2,
        n_limit: int = 20000,
    ):
        """
        Args:
            radius: soft detection radius, kernel size is (2 * radius + 1)
            top_k: top_k > 0: return top k keypoints
            scores_th: top_k <= 0 threshold mode:
                scores_th > 0: return keypoints with scores>scores_th
                else: return keypoints with scores > scores.mean()
            n_limit: max number of keypoint in threshold mode
        """
        super().__init__()
        self.radius = radius
        self.top_k = top_k
        self.scores_th = scores_th
        self.n_limit = n_limit
        self.kernel_size = 2 * self.radius + 1
        self.temperature = 0.1  # tuned temperature
        self.unfold = nn.Unfold(kernel_size=self.kernel_size, padding=self.radius)
        # local xy grid
        x = torch.linspace(-self.radius, self.radius, self.kernel_size)
        # (kernel_size*kernel_size) x 2 : (w,h)
        kw = {"indexing": "ij"} if torch.__version__ >= "1.10" else {}
        self.hw_grid = (
            torch.stack(torch.meshgrid([x, x], **kw)).view(2, -1).t()[:, [1, 0]]
        )

    def forward(
        self,
        scores_map: torch.Tensor,
        sub_pixel: bool = True,
        image_size: Optional[torch.Tensor] = None,
    ):
        """
        :param scores_map: Bx1xHxW
        :param descriptor_map: BxCxHxW
        :param sub_pixel: whether to use sub-pixel keypoint detection
        :return: kpts: list[Nx2,...]; kptscores: list[N,....] normalised position: -1~1
        """
        b, c, h, w = scores_map.shape
        scores_nograd = scores_map.detach()
        nms_scores = simple_nms(scores_nograd, self.radius)

        # remove border
        nms_scores[:, :, : self.radius, :] = 0
        nms_scores[:, :, :, : self.radius] = 0
        if image_size is not None:
            for i in range(scores_map.shape[0]):
                w, h = image_size[i].long()
                nms_scores[i, :, h.item() - self.radius :, :] = 0
                nms_scores[i, :, :, w.item() - self.radius :] = 0
        else:
            nms_scores[:, :, -self.radius :, :] = 0
            nms_scores[:, :, :, -self.radius :] = 0

        # detect keypoints without grad
        if self.top_k > 0:
            topk = torch.topk(nms_scores.view(b, -1), self.top_k)
            indices_keypoints = [topk.indices[i] for i in range(b)]  # B x top_k
        else:
            if self.scores_th > 0:
                masks = nms_scores > self.scores_th
                if masks.sum() == 0:
                    th = scores_nograd.reshape(b, -1).mean(dim=1)  # th = self.scores_th
                    masks = nms_scores > th.reshape(b, 1, 1, 1)
            else:
                th = scores_nograd.reshape(b, -1).mean(dim=1)  # th = self.scores_th
                masks = nms_scores > th.reshape(b, 1, 1, 1)
            masks = masks.reshape(b, -1)

            indices_keypoints = []  # list, B x (any size)
            scores_view = scores_nograd.reshape(b, -1)
            for mask, scores in zip(masks, scores_view):
                indices = mask.nonzero()[:, 0]
                if len(indices) > self.n_limit:
                    kpts_sc = scores[indices]
                    sort_idx = kpts_sc.sort(descending=True)[1]
                    sel_idx = sort_idx[: self.n_limit]
                    indices = indices[sel_idx]
                indices_keypoints.append(indices)

        wh = torch.tensor([w - 1, h - 1], device=scores_nograd.device)

        keypoints = []
        scoredispersitys = []
        kptscores = []
        if sub_pixel:
            # detect soft keypoints with grad backpropagation
            patches = self.unfold(scores_map)  # B x (kernel**2) x (H*W)
            self.hw_grid = self.hw_grid.to(scores_map)  # to device
            for b_idx in range(b):
                patch = patches[b_idx].t()  # (H*W) x (kernel**2)
                indices_kpt = indices_keypoints[
                    b_idx
                ]  # one dimension vector, say its size is M
                patch_scores = patch[indices_kpt]  # M x (kernel**2)
                keypoints_xy_nms = torch.stack(
                    [indices_kpt % w, torch.div(indices_kpt, w, rounding_mode="trunc")],
                    dim=1,
                )  # Mx2

                # max is detached to prevent undesired backprop loops in the graph
                max_v = patch_scores.max(dim=1).values.detach()[:, None]
                x_exp = (
                    (patch_scores - max_v) / self.temperature
                ).exp()  # M * (kernel**2), in [0, 1]

                # \frac{ \sum{(i,j) \times \exp(x/T)} }{ \sum{\exp(x/T)} }
                xy_residual = (
                    x_exp @ self.hw_grid / x_exp.sum(dim=1)[:, None]
                )  # Soft-argmax, Mx2

                hw_grid_dist2 = (
                    torch.norm(
                        (self.hw_grid[None, :, :] - xy_residual[:, None, :])
                        / self.radius,
                        dim=-1,
                    )
                    ** 2
                )
                scoredispersity = (x_exp * hw_grid_dist2).sum(dim=1) / x_exp.sum(dim=1)

                # compute result keypoints
                keypoints_xy = keypoints_xy_nms + xy_residual
                keypoints_xy = keypoints_xy / wh * 2 - 1  # (w,h) -> (-1~1,-1~1)

                kptscore = torch.nn.functional.grid_sample(
                    scores_map[b_idx].unsqueeze(0),
                    keypoints_xy.view(1, 1, -1, 2),
                    mode="bilinear",
                    align_corners=True,
                )[
                    0, 0, 0, :
                ]  # CxN

                keypoints.append(keypoints_xy)
                scoredispersitys.append(scoredispersity)
                kptscores.append(kptscore)
        else:
            for b_idx in range(b):
                indices_kpt = indices_keypoints[
                    b_idx
                ]  # one dimension vector, say its size is M
                # To avoid warning: UserWarning: __floordiv__ is deprecated
                keypoints_xy_nms = torch.stack(
                    [indices_kpt % w, torch.div(indices_kpt, w, rounding_mode="trunc")],
                    dim=1,
                )  # Mx2
                keypoints_xy = keypoints_xy_nms / wh * 2 - 1  # (w,h) -> (-1~1,-1~1)
                kptscore = torch.nn.functional.grid_sample(
                    scores_map[b_idx].unsqueeze(0),
                    keypoints_xy.view(1, 1, -1, 2),
                    mode="bilinear",
                    align_corners=True,
                )[
                    0, 0, 0, :
                ]  # CxN
                keypoints.append(keypoints_xy)
                scoredispersitys.append(kptscore)  # for jit.script compatability
                kptscores.append(kptscore)

        return keypoints, scoredispersitys, kptscores


class InputPadder(object):
    """Pads images such that dimensions are divisible by 8"""

    def __init__(self, h: int, w: int, divis_by: int = 8):
        self.ht = h
        self.wd = w
        pad_ht = (((self.ht // divis_by) + 1) * divis_by - self.ht) % divis_by
        pad_wd = (((self.wd // divis_by) + 1) * divis_by - self.wd) % divis_by
        self._pad = [
            pad_wd // 2,
            pad_wd - pad_wd // 2,
            pad_ht // 2,
            pad_ht - pad_ht // 2,
        ]

    def pad(self, x: torch.Tensor):
        assert x.ndim == 4
        return F.pad(x, self._pad, mode="replicate")

    def unpad(self, x: torch.Tensor):
        assert x.ndim == 4
        ht = x.shape[-2]
        wd = x.shape[-1]
        c = [self._pad[2], ht - self._pad[3], self._pad[0], wd - self._pad[1]]
        return x[..., c[0] : c[1], c[2] : c[3]]


class DeformableConv2d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        stride=1,
        padding=1,
        bias=False,
        mask=False,
    ):
        super(DeformableConv2d, self).__init__()

        self.padding = padding
        self.mask = mask

        self.channel_num = (
            3 * kernel_size * kernel_size if mask else 2 * kernel_size * kernel_size
        )
        self.offset_conv = nn.Conv2d(
            in_channels,
            self.channel_num,
            kernel_size=kernel_size,
            stride=stride,
            padding=self.padding,
            bias=True,
        )

        self.regular_conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=self.padding,
            bias=bias,
        )

    def forward(self, x):
        h, w = x.shape[2:]
        max_offset = max(h, w) / 4.0

        out = self.offset_conv(x)
        if self.mask:
            o1, o2, mask = torch.chunk(out, 3, dim=1)
            offset = torch.cat((o1, o2), dim=1)
            mask = torch.sigmoid(mask)
        else:
            offset = out
            mask = None
        offset = offset.clamp(-max_offset, max_offset)
        x = torchvision.ops.deform_conv2d(
            input=x,
            offset=offset,
            weight=self.regular_conv.weight,
            bias=self.regular_conv.bias,
            padding=self.padding,
            mask=mask,
        )
        return x


def get_conv(
    inplanes,
    planes,
    kernel_size=3,
    stride=1,
    padding=1,
    bias=False,
    conv_type="conv",
    mask=False,
):
    if conv_type == "conv":
        conv = nn.Conv2d(
            inplanes,
            planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
        )
    elif conv_type == "dcn":
        conv = DeformableConv2d(
            inplanes,
            planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=_pair(padding),
            bias=bias,
            mask=mask,
        )
    else:
        raise TypeError
    return conv


class ConvBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        gate: Optional[Callable[..., nn.Module]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        conv_type: str = "conv",
        mask: bool = False,
    ):
        super().__init__()
        if gate is None:
            self.gate = nn.ReLU(inplace=True)
        else:
            self.gate = gate
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self.conv1 = get_conv(
            in_channels, out_channels, kernel_size=3, conv_type=conv_type, mask=mask
        )
        self.bn1 = norm_layer(out_channels)
        self.conv2 = get_conv(
            out_channels, out_channels, kernel_size=3, conv_type=conv_type, mask=mask
        )
        self.bn2 = norm_layer(out_channels)

    def forward(self, x):
        x = self.gate(self.bn1(self.conv1(x)))  # B x in_channels x H x W
        x = self.gate(self.bn2(self.conv2(x)))  # B x out_channels x H x W
        return x


# modified based on torchvision\models\resnet.py#27->BasicBlock
class ResBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        gate: Optional[Callable[..., nn.Module]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        conv_type: str = "conv",
        mask: bool = False,
    ) -> None:
        super(ResBlock, self).__init__()
        if gate is None:
            self.gate = nn.ReLU(inplace=True)
        else:
            self.gate = gate
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("ResBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in ResBlock")
        # Both self.conv1 and self.downsample layers
        # downsample the input when stride != 1
        self.conv1 = get_conv(
            inplanes, planes, kernel_size=3, conv_type=conv_type, mask=mask
        )
        self.bn1 = norm_layer(planes)
        self.conv2 = get_conv(
            planes, planes, kernel_size=3, conv_type=conv_type, mask=mask
        )
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.gate(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.gate(out)

        return out


class SDDH(nn.Module):
    def __init__(
        self,
        dims: int,
        kernel_size: int = 3,
        n_pos: int = 8,
        gate=nn.ReLU(),
        conv2D=False,
        mask=False,
    ):
        super(SDDH, self).__init__()
        self.kernel_size = kernel_size
        self.n_pos = n_pos
        self.conv2D = conv2D
        self.mask = mask

        self.get_patches_func = get_patches

        # estimate offsets
        self.channel_num = 3 * n_pos if mask else 2 * n_pos
        self.offset_conv = nn.Sequential(
            nn.Conv2d(
                dims,
                self.channel_num,
                kernel_size=kernel_size,
                stride=1,
                padding=0,
                bias=True,
            ),
            gate,
            nn.Conv2d(
                self.channel_num,
                self.channel_num,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=True,
            ),
        )

        # sampled feature conv
        self.sf_conv = nn.Conv2d(
            dims, dims, kernel_size=1, stride=1, padding=0, bias=False
        )

        # convM
        if not conv2D:
            # deformable desc weights
            agg_weights = torch.nn.Parameter(torch.rand(n_pos, dims, dims))
            self.register_parameter("agg_weights", agg_weights)
        else:
            self.convM = nn.Conv2d(
                dims * n_pos, dims, kernel_size=1, stride=1, padding=0, bias=False
            )

    def forward(self, x, keypoints):
        # x: [B,C,H,W]
        # keypoints: list, [[N_kpts,2], ...] (w,h)
        b, c, h, w = x.shape
        wh = torch.tensor([[w - 1, h - 1]], device=x.device)
        max_offset = max(h, w) / 4.0

        offsets = []
        descriptors = []
        # get offsets for each keypoint
        for ib in range(b):
            xi, kptsi = x[ib], keypoints[ib]
            kptsi_wh = (kptsi / 2 + 0.5) * wh
            N_kpts = len(kptsi)

            if self.kernel_size > 1:
                patch = self.get_patches_func(
                    xi, kptsi_wh.long(), self.kernel_size
                )  # [N_kpts, C, K, K]
            else:
                kptsi_wh_long = kptsi_wh.long()
                patch = (
                    xi[:, kptsi_wh_long[:, 1], kptsi_wh_long[:, 0]]
                    .permute(1, 0)
                    .reshape(N_kpts, c, 1, 1)
                )

            offset = self.offset_conv(patch).clamp(
                -max_offset, max_offset
            )  # [N_kpts, 2*n_pos, 1, 1]
            if self.mask:
                offset = (
                    offset[:, :, 0, 0].view(N_kpts, 3, self.n_pos).permute(0, 2, 1)
                )  # [N_kpts, n_pos, 3]
                offset = offset[:, :, :-1]  # [N_kpts, n_pos, 2]
                mask_weight = torch.sigmoid(offset[:, :, -1])  # [N_kpts, n_pos]
            else:
                offset = (
                    offset[:, :, 0, 0].view(N_kpts, 2, self.n_pos).permute(0, 2, 1)
                )  # [N_kpts, n_pos, 2]
            offsets.append(offset)  # for visualization

            # get sample positions
            pos = kptsi_wh.unsqueeze(1) + offset  # [N_kpts, n_pos, 2]
            pos = 2.0 * pos / wh[None] - 1
            pos = pos.reshape(1, N_kpts * self.n_pos, 1, 2)

            # sample features
            features = F.grid_sample(
                xi.unsqueeze(0), pos, mode="bilinear", align_corners=True
            )  # [1,C,(N_kpts*n_pos),1]
            features = features.reshape(c, N_kpts, self.n_pos, 1).permute(
                1, 0, 2, 3
            )  # [N_kpts, C, n_pos, 1]
            if self.mask:
                features = torch.einsum("ncpo,np->ncpo", features, mask_weight)

            features = torch.selu_(self.sf_conv(features)).squeeze(
                -1
            )  # [N_kpts, C, n_pos]
            # convM
            if not self.conv2D:
                descs = torch.einsum(
                    "ncp,pcd->nd", features, self.agg_weights
                )  # [N_kpts, C]
            else:
                features = features.reshape(N_kpts, -1)[
                    :, :, None, None
                ]  # [N_kpts, C*n_pos, 1, 1]
                descs = self.convM(features).squeeze()  # [N_kpts, C]

            # normalize
            descs = F.normalize(descs, p=2.0, dim=1)
            descriptors.append(descs)

        return descriptors, offsets


class ALIKED(torch.nn.Module):
    default_conf = {
        "model_name": "aliked-n16",
        "max_num_keypoints": -1,
        "detection_threshold": 0.2,
        "force_num_keypoints": False,
        "pretrained": True,
        "nms_radius": 2,
    }

    checkpoint_url = "https://github.com/Shiaoming/ALIKED/raw/main/models/{}.pth"

    n_limit_max = 20000

    cfgs = {
        "aliked-t16": {
            "c1": 8,
            "c2": 16,
            "c3": 32,
            "c4": 64,
            "dim": 64,
            "K": 3,
            "M": 16,
        },
        "aliked-n16": {
            "c1": 16,
            "c2": 32,
            "c3": 64,
            "c4": 128,
            "dim": 128,
            "K": 3,
            "M": 16,
        },
        "aliked-n": { # == aliked-n16
            "c1": 16,
            "c2": 32,
            "c3": 64,
            "c4": 128,
            "dim": 128,
            "K": 3,
            "M": 16,
        },
        "aliked-n16rot": {
            "c1": 16,
            "c2": 32,
            "c3": 64,
            "c4": 128,
            "dim": 128,
            "K": 3,
            "M": 16,
        },
        "aliked-n32": {
            "c1": 16,
            "c2": 32,
            "c3": 64,
            "c4": 128,
            "dim": 128,
            "K": 3,
            "M": 32,
        },
    }

    required_data_keys = ["image"]

    def __init__(self, conf={}):
        super().__init__()
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        
        if conf.model_name == "aliked-n":
            conf.model_name = "aliked-n16"

        if conf.force_num_keypoints:
            assert conf.detection_threshold <= 0 and conf.max_num_keypoints > 0
        # get configurations
        c1, c2, c3, c4, dim, K, M = [v for _, v in self.cfgs[conf.model_name].items()]
        conv_types = ["conv", "conv", "dcn", "dcn"]
        conv2D = False
        mask = False

        # build model
        self.pool2 = nn.AvgPool2d(kernel_size=2, stride=2)
        self.pool4 = nn.AvgPool2d(kernel_size=4, stride=4)
        self.norm = nn.BatchNorm2d
        self.gate = nn.SELU(inplace=True)
        self.block1 = ConvBlock(3, c1, self.gate, self.norm, conv_type=conv_types[0])
        self.block2 = ResBlock(
            c1,
            c2,
            1,
            nn.Conv2d(c1, c2, 1),
            gate=self.gate,
            norm_layer=self.norm,
            conv_type=conv_types[1],
        )
        self.block3 = ResBlock(
            c2,
            c3,
            1,
            nn.Conv2d(c2, c3, 1),
            gate=self.gate,
            norm_layer=self.norm,
            conv_type=conv_types[2],
            mask=mask,
        )
        self.block4 = ResBlock(
            c3,
            c4,
            1,
            nn.Conv2d(c3, c4, 1),
            gate=self.gate,
            norm_layer=self.norm,
            conv_type=conv_types[3],
            mask=mask,
        )
        self.conv1 = resnet.conv1x1(c1, dim // 4)
        self.conv2 = resnet.conv1x1(c2, dim // 4)
        self.conv3 = resnet.conv1x1(c3, dim // 4)
        self.conv4 = resnet.conv1x1(dim, dim // 4)
        self.upsample2 = nn.Upsample(
            scale_factor=2, mode="bilinear", align_corners=True
        )
        self.upsample4 = nn.Upsample(
            scale_factor=4, mode="bilinear", align_corners=True
        )
        self.upsample8 = nn.Upsample(
            scale_factor=8, mode="bilinear", align_corners=True
        )
        self.upsample32 = nn.Upsample(
            scale_factor=32, mode="bilinear", align_corners=True
        )
        self.score_head = nn.Sequential(
            resnet.conv1x1(dim, 8),
            self.gate,
            resnet.conv3x3(8, 4),
            self.gate,
            resnet.conv3x3(4, 4),
            self.gate,
            resnet.conv3x3(4, 1),
        )
        self.desc_head = SDDH(dim, K, M, gate=self.gate, conv2D=conv2D, mask=mask)
        self.dkd = DKD(
            radius=conf.nms_radius,
            top_k=-1 if conf.detection_threshold > 0 else conf.max_num_keypoints,
            scores_th=conf.detection_threshold,
            n_limit=conf.max_num_keypoints
            if conf.max_num_keypoints > 0
            else self.n_limit_max,
        )

        # load pretrained
        if conf.pretrained:
            state_dict = torch.hub.load_state_dict_from_url(
                self.checkpoint_url.format(conf.model_name), map_location="cpu"
            )
            self.load_state_dict(state_dict, strict=True)

    def extract_dense_map(self, image):
        # Pads images such that dimensions are divisible by
        div_by = 2**5
        padder = InputPadder(image.shape[-2], image.shape[-1], div_by)
        image = padder.pad(image)

        # ================================== feature encoder
        x1 = self.block1(image)  # B x c1 x H x W
        x2 = self.pool2(x1)
        x2 = self.block2(x2)  # B x c2 x H/2 x W/2
        x3 = self.pool4(x2)
        x3 = self.block3(x3)  # B x c3 x H/8 x W/8
        x4 = self.pool4(x3)
        x4 = self.block4(x4)  # B x dim x H/32 x W/32
        # ================================== feature aggregation
        x1 = self.gate(self.conv1(x1))  # B x dim//4 x H x W
        x2 = self.gate(self.conv2(x2))  # B x dim//4 x H//2 x W//2
        x3 = self.gate(self.conv3(x3))  # B x dim//4 x H//8 x W//8
        x4 = self.gate(self.conv4(x4))  # B x dim//4 x H//32 x W//32
        x2_up = self.upsample2(x2)  # B x dim//4 x H x W
        x3_up = self.upsample8(x3)  # B x dim//4 x H x W
        x4_up = self.upsample32(x4)  # B x dim//4 x H x W
        x1234 = torch.cat([x1, x2_up, x3_up, x4_up], dim=1)
        # ================================== score head
        score_map = torch.sigmoid(self.score_head(x1234))
        feature_map = torch.nn.functional.normalize(x1234, p=2, dim=1)

        # Unpads images
        feature_map = padder.unpad(feature_map)
        score_map = padder.unpad(score_map)

        return feature_map, score_map

    def forward(self, data):
        image = data["image"]
        feature_map, score_map = self.extract_dense_map(image)
        keypoints, kptscores, scoredispersitys = self.dkd(
            score_map, image_size=data.get("image_size")
        )
        descriptors, offsets = self.desc_head(feature_map, keypoints)

        _, _, h, w = image.shape
        wh = torch.tensor([w, h], device=image.device)
        # no padding required,
        # we can set detection_threshold=-1 and conf.max_num_keypoints
        return {
            "keypoints": wh * (torch.stack(keypoints) + 1) / 2.0,  # B N 2
            "descriptors": torch.stack(descriptors),  # B N D
            "keypoint_scores": torch.stack(kptscores),  # B N
            "score_dispersity": torch.stack(scoredispersitys),
            "score_map": score_map,  # Bx1xHxW
        }

    def forward_desc(self, data, keypoints):
        image = data["image"]
        b, _, h, w = image.shape
        wh = torch.tensor([w, h], device=image.device)
        
        # keypoints: BxNx2 (w,h) -> BxNx2 (-1+1)
        keypoints = 2.0 * keypoints / wh - 1

        feature_map, scoremap = self.extract_dense_map(image)
        descriptors, offsets = self.desc_head(feature_map, keypoints)
        scores = torch.nn.functional.grid_sample(
            scoremap, keypoints.view(b, -1, 1, 2), mode="bilinear", align_corners=True
        ).squeeze()
        
        keypoints = (keypoints + 1) / 2 * wh  # BxNx2 (-1+1) -> BxNx2 (w,h)
        
        return {
            "descriptors": torch.stack(descriptors),  # B N D
            "keypoints": keypoints,  # B N 2
            "keypoint_scores": scores,  # B N
        }
        

    


import torch.nn as nn


class Registry(object):

    def __init__(self, name):
        self._name = name
        self._module_dict = dict()

    @property
    def name(self):
        return self._name

    @property
    def module_dict(self):
        return self._module_dict

    def _register_module(self, module_class):
        """Register a module.

        Args:
            module (:obj:`nn.Module`): Module to be registered.
        """
        if not issubclass(module_class, nn.Module):
            raise TypeError(
                'module must be a child of nn.Module, but got {}'.format(
                    module_class))
        module_name = module_class.__name__
        if module_name in self._module_dict:
            raise KeyError('{} is already registered in {}'.format(
                module_name, self.name))
        self._module_dict[module_name] = module_class

    def register_module(self, cls):
        self._register_module(cls)
        return cls


BACKBONES = Registry('backbone')
NECKS = Registry('neck')
ROI_EXTRACTORS = Registry('roi_extractor')
SHARED_HEADS = Registry('shared_head')
HEADS = Registry('head')
LOSSES = Registry('loss')
DETECTORS = Registry('detector')


# -*- coding: UTF-8 -*-
'''=================================================
@Project -> File   r2d2 -> extract_localization
@IDE    PyCharm
@Author fx221@cam.ac.uk
@Date   13/07/2022 09:59
=================================================='''
import os
import os.path as osp
import h5py
import numpy as np
import cv2
import torch.utils.data as Data
from tqdm import tqdm
from types import SimpleNamespace
import logging
import pprint
from pathlib import Path

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as tvf

from easy_local_features.submodules.git_sfd2.nets import ResSegNet, ResSegNetV2
from easy_local_features.submodules.git_sfd2.utils import extract_resnet_return

class ImageDataset(Data.Dataset):
    default_conf = {
        'globs': ['*.jpg', '*.png', '*.jpeg', '*.JPG', '*.PNG'],
        'grayscale': False,
        'resize_max': None,
        'resize_force': False,
    }

    def __init__(self, root, conf, image_list=None,
                 mask_root=None):
        self.conf = conf = SimpleNamespace(**{**self.default_conf, **conf})
        self.root = root

        self.paths = []
        if image_list is None:
            for g in conf.globs:
                self.paths += list(Path(root).glob('**/' + g))
            if len(self.paths) == 0:
                raise ValueError(f'Could not find any image in root: {root}.')
            self.paths = [i.relative_to(root) for i in self.paths]
        else:
            with open(image_list, "r") as f:
                lines = f.readlines()
                for l in lines:
                    l = l.strip()
                    self.paths.append(Path(l))

        logging.info(f'Found {len(self.paths)} images in root {root}.')

        if mask_root is not None:
            self.mask_root = mask_root
        else:
            self.mask_root = None

        print("mask_root: ", self.mask_root)

    def __getitem__(self, idx):
        path = self.paths[idx]
        if self.conf.grayscale:
            mode = cv2.IMREAD_GRAYSCALE
        else:
            mode = cv2.IMREAD_COLOR
        image = cv2.imread(str(self.root / path), mode)
        if not self.conf.grayscale:
            image = image[:, :, ::-1]  # BGR to RGB
        if image is None:
            raise ValueError(f'Cannot read image {str(path)}.')
        image = image.astype(np.float32)
        size = image.shape[:2][::-1]
        w, h = size

        if self.conf.resize_max and (self.conf.resize_force
                                     or max(w, h) > self.conf.resize_max):
            scale = self.conf.resize_max / max(h, w)
            h_new, w_new = int(round(h * scale)), int(round(w * scale))
            image = cv2.resize(
                image, (w_new, h_new), interpolation=cv2.INTER_CUBIC)

        if self.conf.grayscale:
            image = image[None]
        else:
            image = image.transpose((2, 0, 1))  # HxWxC to CxHxW
        image = image / 255.

        data = {
            'name': str(path),
            'image': image,
            'original_size': np.array(size),
        }

        if self.mask_root is not None:
            mask_path = Path(str(path).replace("jpg", "png"))
            if osp.exists(mask_path):
                mask = cv2.imread(str(self.mask_root / mask_path))
                mask = cv2.resize(mask, dsize=(image.shape[2], image.shape[1]), interpolation=cv2.INTER_NEAREST)
            else:
                mask = np.zeros(shape=(image.shape[1], image.shape[2], 3), dtype=np.uint8)

            data['mask'] = mask

        return data

    def __len__(self):
        return len(self.paths)


def get_model(model_name, use_stability=False):
    if model_name == 'ressegnet':
        model = ResSegNet(outdim=128, require_stability=use_stability).eval()
        # model.load_state_dict(torch.load(weight_path)['model'], strict=True)
        extractor = extract_resnet_return
    if model_name == 'ressegnetv2':
        model = ResSegNetV2(outdim=128, require_stability=use_stability).eval()
        # model.load_state_dict(torch.load(weight_path)['model'], strict=False)
        extractor = extract_resnet_return

    return model, extractor

class SFD2(nn.Module):
    def __init__(self, 
            model_name="ressegnetv2", # "ressegnetv2", "ressegnet"
            use_stability=True,
            top_k=4096,
            conf_th=0.001,
            scales=[1.0],
        ):
        super().__init__()
        self.model, self.extractor = get_model(model_name, use_stability)
        self.top_k = top_k
        self.conf_th = conf_th
        self.scales = scales
        self.model.eval()
        
    def forward(self, data):
        pred = self.extractor(self.model, img=data["image"],
                    topK=self.top_k,
                    mask=None,
                    conf_th=self.conf_th,
                    scales=self.scales,
                )

        # pred['image_size'] = original_size = data['original_size'][0].cpu().numpy()
        # if 'keypoints' in pred.keys():
            # size = np.array(data['image'].shape[-2:][::-1])
            # scales = (original_size / size).astype(np.float32)
            # pred['keypoints'] = (pred['keypoints'] + .5) * scales[None] - .5
            
        return pred



from matplotlib.patches import ConnectionPatch
import matplotlib.pyplot as plt
import numpy as np
import torch
import torchvision
import os

from .ops import to_cv, sort_keypoints

default_colors = {
    'g': '#4ade80',
    'r': '#ef4444',
    'b': '#3b82f6',
    'y': '#fbbf24',  # yellow
    'p': '#a855f7',  # purple
    'o': '#fb923c',  # orange
    'c': '#06b6d4',  # cyan
    'm': '#ec4899',  # magenta
}

def draw_keypoints(image, keypoints, color='g', size=5):
    print(keypoints.shape)
    # draw keypoints on the image
    imgs = []
    for B in range(keypoints.shape[0]):
        kps = keypoints[B].unsqueeze(0)
        im = image[B]
        
        if color in default_colors:
            col = default_colors[color]
        else:
            col = color
        
        im = torchvision.utils.draw_keypoints(im, kps, radius=size, colors=col)
        
        imgs.append(im)
    return torch.stack(imgs)
        

def plot_pair(img0, img1, figsize=(20, 10), fig=None, ax=None, title=None, vertical=False, gray=True) -> tuple[plt.Figure, list[plt.Axes]]:
    if fig is None:
        if vertical:
            fig, ax = plt.subplots(2, 1, figsize=figsize)
        else:
            fig, ax = plt.subplots(1, 2, figsize=figsize)
            
    if gray:
        ax[0].imshow(to_cv(img0, to_gray=gray), cmap='gray')
        ax[1].imshow(to_cv(img1, to_gray=gray), cmap='gray')
    else:
        ax[0].imshow(to_cv(img0))
        ax[1].imshow(to_cv(img1))
    
    # remove border
    for a in ax:
        a.axis('off')
    # remove space between subplots
    plt.subplots_adjust(wspace=0, hspace=0)
    
    if title is not None:
        fig.suptitle(title)
    return fig, ax

def plot_image(image, figsize=(10, 10), fig=None, ax=None, title=None, gray=True) -> tuple[plt.Figure, list[plt.Axes]]:
    if fig is None or ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    
    if gray:
        ax.imshow(to_cv(image, to_gray=gray), cmap='gray')
    else:
        ax.imshow(to_cv(image))
    
    ax.axis('off')
    
    if title is not None:
        fig.suptitle(title)
    return fig, ax

def plot_depth_pair(img0, img1, fig=None, ax=None, title=None) -> tuple[plt.Figure, list[plt.Axes]]:
    if fig is None or ax is None:
        fig = plt.gcf()
        ax = fig.axes

    ax[0].imshow(to_cv(img0, to_gray=True), cmap='jet', alpha=0.5)
    ax[1].imshow(to_cv(img1, to_gray=True), cmap='jet', alpha=0.5)
    
    if title is not None:
        fig.suptitle(title)
    return fig, ax

def plot_depth(depth, fig=None, ax=None, title=None) -> tuple[plt.Figure, list[plt.Axes]]:
    if fig is None or ax is None:
        fig = plt.gcf()
        ax = fig.axes

    ax[0].imshow(depth, cmap='jet', alpha=0.5)
    
    if title is not None:
        fig.suptitle(title)
    return fig, ax

def plot_keypoints(keypoints0=None, keypoints1=None, fig=None, ax=None, color=None, kps_size=5, all_colors=None, **kwargs) -> tuple[plt.Figure, list[plt.Axes]]:
    rainbow = plt.get_cmap('hsv')
    if fig is None or ax is None:
        fig = plt.gcf()
        ax = fig.axes
        
    # if we have both keypoints sort them by x and then by y
    if keypoints0 is not None and keypoints1 is not None:
        keypoints0, keypoints1 = sort_keypoints(keypoints0, keypoints1)
    
    if keypoints0 is not None:        
        if all_colors is None:
            if isinstance(color, str):
                if color in default_colors:
                    col = default_colors[color]
                else:
                    col = color
                all_colors0 = [col for _ in range(len(keypoints0))]
            else:
                all_colors0 = [rainbow(i / len(keypoints0)) for i in range(len(keypoints0))]
        else:
            all_colors0 = all_colors

        if isinstance(keypoints0, torch.Tensor):
            keypoints0 = keypoints0.detach().cpu().numpy()
        if len(keypoints0.shape) == 3:
            keypoints0 = keypoints0.squeeze(0)
            
        ax[0].scatter(keypoints0[:, 0], keypoints0[:, 1], s=kps_size, c=all_colors0, **kwargs)

    if keypoints1 is not None:
        if all_colors is None:
            if isinstance(color, str):
                if color in default_colors:
                    col = default_colors[color]
                else:
                    col = color
                all_colors1 = [col for _ in range(len(keypoints1))]
            else:
                all_colors1 = [rainbow(i / len(keypoints1)) for i in range(len(keypoints1))]
        else:
            all_colors1 = all_colors
            
        if isinstance(keypoints1, torch.Tensor):
            keypoints1 = keypoints1.detach().cpu().numpy()
        if len(keypoints1.shape) == 3:
            keypoints1 = keypoints1.squeeze(0)
        ax[1].scatter(keypoints1[:, 0], keypoints1[:, 1], s=kps_size, c=all_colors1, **kwargs)

    return fig, ax

def plot_matches(mkpts0, mkpts1, fig=None, ax=None, color=None, **kwargs):
    if fig is None or ax is None:
        fig = plt.gcf()
        ax = fig.axes
    
    if (isinstance(color, str)) and (color in default_colors):
        color = default_colors[color]
    
    if isinstance(mkpts0, torch.Tensor):
        mkpts0 = mkpts0.detach().cpu().numpy()
    if isinstance(mkpts1, torch.Tensor):
        mkpts1 = mkpts1.detach().cpu().numpy()
    
    if color is None:
        rainbow = plt.get_cmap('hsv')
        color = [rainbow(i / len(mkpts0)) for i in range(len(mkpts0))]
    
    if not isinstance(color, list):
        color = [color for _ in range(len(mkpts0))]
        
    # if we have the color in the default colors, replace it
    for i, c in enumerate(color):
        if c in default_colors:
            color[i] = default_colors[c]
            
    mkpts0, mkpts1 = sort_keypoints(mkpts0, mkpts1)
    
    for i, (mkp0, mkp1) in enumerate(zip(mkpts0, mkpts1)):
        con = ConnectionPatch(
            xyA=mkp0, xyB=mkp1, 
            coordsA="data", coordsB="data",
            axesA=ax[0], axesB=ax[1], 
            color=color[i], linewidth=0.5, **kwargs)
        con.set_in_layout(False) # remove from layout calculations
        ax[0].add_artist(con)
    ax[1].set_zorder(-1)
    
    return fig, ax

def add_text(text, fig=None, ax=None, **kwargs):
    if fig is None or ax is None:
        fig = plt.gcf()
        ax = fig.axes[0]
    ax.text(0, 0, text, color='black', fontsize=12, ha='left', va='top', bbox=dict(facecolor='white', alpha=0.5, edgecolor='white'))
    return fig, ax

def show():
    plt.tight_layout()
    plt.show()
    
def save(path):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    
    plt.tight_layout()
    plt.savefig(path)
    plt.close()


import os, sys
import glob
import json
import numpy as np

import fnmatch
import os

import tqdm

DEFAULT_DATA_PATH = '/srv/storage/datasets/cadar/simulator/_dataset_test2/' 
OUTPUT_PATH = '/srv/storage/datasets/cadar/semantic-local-features/metrics_output/simulation/brand/'
SIFT_PATH = '/srv/storage/datasets/cadar/semantic-local-features/metrics_output/simulation/sift/'
BIN = '/srv/storage/datasets/cadar/easy-local-features-baselines/easy_local_features/submodules/git_brand/build/desc_brand'

if __name__ == "__main__":
    # images = glob.glob(DEFAULT_DATA_PATH + '/**/rgba*.png', recursive=True)
    # desc_brand fx fy cx cy n <rgb_image depth_image keypoints_csv output_file>

    images = []
    for root, dirnames, filenames in os.walk(DEFAULT_DATA_PATH):
        for filename in fnmatch.filter(filenames, 'rgba*.png'):
            images.append(os.path.join(root, filename))

    if not os.path.exists(OUTPUT_PATH):
        os.makedirs(OUTPUT_PATH)

    metadata_file = images[0].split('rgba')[0] + 'metadata.json'
    metadata = json.load(open(metadata_file, 'r'))
    resolution = metadata['metadata']['resolution']
    focal = metadata['camera']['focal_length']
    sensor_width = metadata['camera']['sensor_width']
    field_of_view = metadata['camera']['field_of_view']

    # print(camera_matrix)

    fx = focal * resolution[0] / sensor_width
    fy = focal * resolution[1] / sensor_width
    cx = resolution[0] / 2
    cy = resolution[1] / 2

    for image_path in tqdm.tqdm(images):
        cmd = "{} {} {} {} {} {} ".format(BIN, fx, fy, cx, cy, 1)
    
        depth_path = image_path.replace('rgba', 'depth').replace('png', 'tiff')
        npz_keypoints_path = image_path.replace(DEFAULT_DATA_PATH, SIFT_PATH) + ".kpts.npz"
        csv_keypoints_path = image_path.replace('rgba', 'sift').replace('.png', '.csv').replace(DEFAULT_DATA_PATH, OUTPUT_PATH)
        output_path = image_path.replace('rgba', 'brand').replace('.png', '.csv').replace(DEFAULT_DATA_PATH, OUTPUT_PATH)

        npz_output_path = image_path.replace(DEFAULT_DATA_PATH, OUTPUT_PATH) + '.desc'

        out_dir = os.path.dirname(output_path)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        # read and turn to csv if not already
        if not os.path.exists(csv_keypoints_path):
            kps = np.load(npz_keypoints_path)['keypoints']
            # save csv keypoints with ", "
            np.savetxt(csv_keypoints_path, kps, delimiter=", ", fmt='%.3f')

            # cp npz_keypoints_path to csv_keypoints_path folder
            dst_folder = os.path.dirname(csv_keypoints_path)
            os.system("cp {} {}".format(npz_keypoints_path, dst_folder))

        cmd += "{} {} {} {}".format(image_path, depth_path, csv_keypoints_path, output_path)

        ret = os.system(cmd)
        if ret != 0:
            print("Error in {}".format(image_path))
            continue

        #read output_path
        descs = np.loadtxt(output_path, delimiter=",", skiprows=1)
        # save as npz
        np.savez_compressed(npz_output_path, descriptors=descs)

        # save the 







import mmcv
from torch import nn

from .registry import (BACKBONES, NECKS, ROI_EXTRACTORS, SHARED_HEADS, HEADS,
                       LOSSES, DETECTORS)


def _build_module(cfg, registry, default_args):
    assert isinstance(cfg, dict) and 'type' in cfg
    assert isinstance(default_args, dict) or default_args is None
    args = cfg.copy()
    obj_type = args.pop('type')
    if mmcv.is_str(obj_type):
        if obj_type not in registry.module_dict:
            raise KeyError('{} is not in the {} registry'.format(
                obj_type, registry.name))
        obj_type = registry.module_dict[obj_type]
    elif not isinstance(obj_type, type):
        raise TypeError('type must be a str or valid type, but got {}'.format(
            type(obj_type)))
    if default_args is not None:
        for name, value in default_args.items():
            args.setdefault(name, value)
    return obj_type(**args)


def build(cfg, registry, default_args=None):
    if isinstance(cfg, list):
        modules = [_build_module(cfg_, registry, default_args) for cfg_ in cfg]
        return nn.Sequential(*modules)
    else:
        return _build_module(cfg, registry, default_args)


def build_backbone(cfg):
    return build(cfg, BACKBONES)


def build_neck(cfg):
    return build(cfg, NECKS)


def build_roi_extractor(cfg):
    return build(cfg, ROI_EXTRACTORS)


def build_shared_head(cfg):
    return build(cfg, SHARED_HEADS)


def build_head(cfg):
    return build(cfg, HEADS)


def build_loss(cfg):
    return build(cfg, LOSSES)


def build_detector(cfg, train_cfg=None, test_cfg=None):
    return build(cfg, DETECTORS, dict(train_cfg=train_cfg, test_cfg=test_cfg))


import torch
import torch.nn.functional as F
from omegaconf import OmegaConf

@torch.no_grad()
def find_nn(sim, ratio_thresh, distance_thresh):
    sim_nn, ind_nn = sim.topk(2 if ratio_thresh else 1, dim=-1, largest=True)
    dist_nn = 2 * (1 - sim_nn)
    mask = torch.ones(ind_nn.shape[:-1], dtype=torch.bool, device=sim.device)
    if ratio_thresh:
        mask = mask & (dist_nn[..., 0] <= (ratio_thresh**2) * dist_nn[..., 1])
    if distance_thresh:
        mask = mask & (dist_nn[..., 0] <= distance_thresh**2)
    matches = torch.where(mask, ind_nn[..., 0], ind_nn.new_tensor(-1))
    return matches

def mutual_check(m0, m1):
    inds0 = torch.arange(m0.shape[-1], device=m0.device)
    inds1 = torch.arange(m1.shape[-1], device=m1.device)
    loop0 = torch.gather(m1, -1, torch.where(m0 > -1, m0, m0.new_tensor(0)))
    loop1 = torch.gather(m0, -1, torch.where(m1 > -1, m1, m1.new_tensor(0)))
    m0_new = torch.where((m0 > -1) & (inds0 == loop0), m0, m0.new_tensor(-1))
    m1_new = torch.where((m1 > -1) & (inds1 == loop1), m1, m1.new_tensor(-1))
    return m0_new, m1_new

def nn_matcher(desc1, desc2, ratio_thresh=None, do_distance_thresh=None, do_mutual_check=True):
    sim = torch.einsum("bnd,bmd->bnm", desc1, desc2)
    matches0 = find_nn(sim, ratio_thresh, do_distance_thresh)
    matches1 = find_nn(sim.transpose(1, 2), ratio_thresh, do_distance_thresh)
    if do_mutual_check:
        matches0, matches1 = mutual_check(matches0, matches1)
    mscores0 = (matches0 > -1).float()
    mscores1 = (matches1 > -1).float()
    return {
        "matches0": matches0,
        "matches1": matches1,
        "matching_scores0": mscores0,
        "matching_scores1": mscores1,
        "similarity": sim,
    }

class NearestNeighborMatcher(torch.nn.Module):
    default_conf = {
        "ratio_thresh": None,
        "distance_thresh": None,
        "mutual_check": True,
        "loss": None,
        "normalize": True,
    }
    required_data_keys = ["descriptors0", "descriptors1"]

    def __init__(self, conf={}):
        super().__init__()
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)

    def forward(self, data):
        if self.conf.normalize:
            data["descriptors0"] = F.normalize(data["descriptors0"], p=2, dim=-1)
            data["descriptors1"] = F.normalize(data["descriptors1"], p=2, dim=-1)
        
        sim = torch.einsum("bnd,bmd->bnm", data["descriptors0"], data["descriptors1"])
        matches0 = find_nn(sim, self.conf.ratio_thresh, self.conf.distance_thresh)
        matches1 = find_nn(
            sim.transpose(1, 2), self.conf.ratio_thresh, self.conf.distance_thresh
        )
        if self.conf.mutual_check:
            matches0, matches1 = mutual_check(matches0, matches1)
        b, m, n = sim.shape
        mscores0 = (matches0 > -1).float()
        mscores1 = (matches1 > -1).float()
        
        return {
            "matches0": matches0,
            "matches1": matches1,
            "matching_scores0": mscores0,
            "matching_scores1": mscores1,
            "similarity": sim,
        }
        
        
class MultipleNearestNeighborMatcher(torch.nn.Module):
    default_conf = {
        "ratio_thresh": None,
        "distance_thresh": None,
        "mutual_check": True,
        "matching_keys": ["semantic_features", "texture_features"],
        "fusion": "product", # "product" or "sum" or "concat"
        "sum_weights": [1, 1],
    }

    def __init__(self, conf):
        super().__init__()
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)

    @torch.no_grad()
    def forward(self, data):
        sims = []
        # normalize all
        for key in self.conf.matching_keys:
            data[key + "0"] = F.normalize(data[key + "0"], p=2, dim=-1)
            data[key + "1"] = F.normalize(data[key + "1"], p=2, dim=-1)
        
        if self.conf.fusion in ["product", "sum"]:
            for key in self.conf.matching_keys:
                sim = torch.einsum("bnd,bmd->bnm", data[key + "0"], data[key + "1"])
                sims.append(sim)
            if self.conf.fusion == "sum":
                sims = [s * w for s, w in zip(sims, self.conf.sum_weights)]
                sim = torch.sum(torch.stack(sims), dim=0)
            else:
                sim = torch.prod(torch.stack(sims), dim=0)
        elif self.conf.fusion == "concat":
            d0 = torch.cat([data[key + "0"] for key in self.conf.matching_keys], dim=-1)
            d1 = torch.cat([data[key + "1"] for key in self.conf.matching_keys], dim=-1)
            sim = torch.einsum("bnd,bmd->bnm", d0, d1)
        
        matches0 = find_nn(sim, self.conf.ratio_thresh, self.conf.distance_thresh)
        matches1 = find_nn(
            sim.transpose(1, 2), self.conf.ratio_thresh, self.conf.distance_thresh
        )
        if self.conf.mutual_check:
            matches0, matches1 = mutual_check(matches0, matches1)
        b, m, n = sim.shape
        # mscores0 = (matches0 > -1).float()
        # mscores1 = (matches1 > -1).float()
        
        
        # pull from similiarity matrix
        mscores0 = sim[torch.arange(b)[:, None], torch.arange(m)[None, :], matches0]
        mscores1 = sim[torch.arange(b)[:, None], matches1, torch.arange(n)[None, :]]
        
        return {
            "matches0": matches0,
            "matches1": matches1,
            "matching_scores0": mscores0,
            "matching_scores1": mscores1,
            "similarity": sim,
        }


import torch
import torch.nn as nn

from .group_align import *


class DescGroupPoolandNorm:
    def __init__(self, args):
        self.pool = 'shift'
        self.norm = "l2"
        self.candidate = args.candidate

        self.G = args.num_group
        self.C = 2*args.channels ## because the aggregation of the multi-layer features is concatenation.


    def desc_pool_and_norm_infer(self, kpts, desc):
        """ kpts [B, K, 2], desc torch.tensor([B, K, CG])
        """
        B, K, CG = desc.shape
        desc = desc.reshape(B,K,-1,self.G)

        if self.pool == 'shift':
            if 'top' in self.candidate:
                topk = int(self.candidate[3])
                kpts_update, desc_update = shift_topk_candidate(kpts, desc, B, K, self.G, topk)
            else:  ## ratio
                ratio = float(self.candidate)
                kpts_update, desc_update = shift_ratio_candidate(kpts, desc, B, K, CG, self.G, ratio)
        else:
            kpts_update = kpts
            desc_update = self._descriptor_pool(desc)

        if 'top' in self.candidate or self.pool!='shift': ## batchified version
            desc_update = self._descriptor_norm(desc_update)
        else: ## ratio (batch as list)
            desc_update_list = []
            for dd in desc_update:
                desc_update_list.append( self._descriptor_norm(dd))
            desc_update = desc_update_list
        
        return kpts_update, desc_update

    def desc_pool_and_norm(self, desc, GTShift):
        """ desc [B, K, CG], for training.
        """

        B, K, CG = desc.shape
        desc = desc.reshape(B,K,-1,self.G)

        if self.pool=='shift':
            shift = GTShift.unsqueeze(1).repeat(1, K)       
            desc_update = shifting(desc, shift, B, K, self.G)
        else:
            desc_update = self._descriptor_pool(desc)

        desc_update = self._descriptor_norm(desc_update)

        self._exception_handling(desc_update, desc, B, K, CG, shift)

        return desc_update

    def _descriptor_pool(self, desc):
        if self.pool=='avg':
            desc_update = desc.mean(dim=3)
        elif self.pool=='weight_avg':
            weight = desc[:, :, 0, :].unsqueeze(2) 
            desc_update = (weight * desc).mean(dim=3)
        elif self.pool=='max':
            desc_update = desc.max(dim=3)[0]
        elif self.pool=='norm':
            desc_update = desc.norm(dim=3)
        elif self.pool=='none':
            desc_update = desc.reshape(B, K, CG)
        else:
            raise NotImplementedError
        return desc_update

    def _descriptor_norm(self, desc_update):
        ## descriptor normalize; default: l2 normalize
        if self.norm=='l2':
            desc_update = L2Norm()(desc_update)
        elif self.norm =='l1':
            desc_update = L1Norm()(desc_update)
        elif self.norm =='context':
            desc_update = ContextNorm()(desc_update)
        elif self.norm == 'none':
            desc_update = desc_update
        else:
            raise NotImplementedError
        return desc_update


    def _exception_handling(self, desc_update, desc, B, K, CG, shift):
        if self.pool == 'shift' or self.pool =='shift_avg' or self.pool =='shift_all':
            assert desc_update.shape == desc.reshape(B,K,CG).shape
            assert shift.max() < self.G and shift.min() >=0
        elif  self.pool =='none':
            assert desc_update.shape == desc.reshape(B,K,CG).shape
        elif self.pool == 'bilinear':
            assert desc_update.shape[-1] == 128
        else:
            assert desc_update.shape[:3] == desc.shape[:3]  ## [B, K, C]


######## descriptor normalize

class L2Norm(nn.Module):
    def __init__(self):
        super(L2Norm, self).__init__()
        self.eps = 1e-6

    def forward(self, x):
        """ x [B, K, CG] or [K, CG]"""
        # norm = torch.sqrt(torch.sum(x * x, dim=1) + self.eps)
        norm = torch.norm(x, p=2, dim=-1) + self.eps
        x = x / norm.unsqueeze(-1)
        return x

class L1Norm(nn.Module):
    def __init__(self):
        super(L1Norm, self).__init__()
        self.eps = 1e-6

    def forward(self, x):
        # norm = torch.sum(torch.abs(x), dim=1) + self.eps
        norm = torch.norm(x, p=1, dim=-1) + self.eps
        x = x / norm.unsqueeze(-1)
        return x

class ContextNorm(nn.Module):
    def __init__(self):
        super(ContextNorm, self).__init__()
        self.eps = 1e-6
    
    def forward(self, x):
        mean = torch.mean(x, dim=-1) 
        std = torch.std(x, dim=-1) + self.eps
        x = (x - mean.unsqueeze(-1)) / std.unsqueeze(-1)
        return x


# logpolar_weight = 'https://github.com/cvlab-epfl/log-polar-descriptors/raw/master/weights/log-polar.pth'
# cartesian_weight = 'https://github.com/cvlab-epfl/log-polar-descriptors/raw/master/weights/cartesian.pth'


# import torch
# import numpy as np
# import torch.nn.functional as F
# from functools import partial
# import cv2, os, wget

# from easy_local_features.submodules.git_logpolar.models import HardNet
# from easy_local_features.submodules.git_logpolar.defaults import _C as cfg

# from ..matching.nearest_neighbor import NearestNeighborMatcher
# from omegaconf import OmegaConf
# from .basemodel import BaseExtractor
# from ..utils.download import downloadModel
# from ..utils import ops

# class LogPolar_baseline(BaseExtractor):
#     default_conf = {
#         'use_log_polar': True,
#     }
    
#     def __init__(self, conf={}):
#         self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        
#         self.DEV   = torch.device('cpu')
#         use_log_polar = conf.use_log_polar
#         this_file = os.path.abspath(__file__)
#         this_dir = os.path.dirname(this_file)

#         if use_log_polar:
#             config_path = os.path.join(this_dir, '../submodules/git_logpolar/init_one_example_ptn_96.yml')
#             weights = logpolar_weight
#         else:
#             config_path = os.path.join(this_dir, '../submodules/git_logpolar/init_one_example_stn_16.yml')
#             weights = cartesian_weight

#         cfg.merge_from_file(config_path)

#         self.model = HardNet(transform=cfg.TEST.TRANSFORMER,
#                     coords=cfg.TEST.COORDS,
#                     patch_size=cfg.TEST.IMAGE_SIZE,
#                     scale=cfg.TEST.SCALE,
#                     is_desc256=cfg.TEST.IS_DESC_256,
#                     orientCorrect=cfg.TEST.ORIENT_CORRECTION)

#         state_dict = torch.hub.load_state_dict_from_url(weights, progress=True)['state_dict']
#         self.model.load_state_dict(state_dict)
#         self.model.to(self.DEV)

#     def detectAndCompute(self, img, return_dict=False):
#         raise NotImplemented

#     def detect(self, img):
#         raise NotImplemented

#     def compute(self, img, kps, scales, orientations, return_dict=False):
#         # make sure image is gray
#         if len(img.shape) == 3:
#             img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

#         pts = np.array([kp.pt for kp in cv_kps])
#         scales = np.array([kp.size for kp in cv_kps])
#         oris = np.array([kp.angle for kp in cv_kps])

#         # Mirror-pad the image to avoid boundary effects
#         if any([s > cfg.TEST.PAD_TO for s in img.shape[:2]]):
#             raise RuntimeError(
#                 "Image exceeds acceptable size ({}x{}), please downsample".format(
#                     cfg.TEST.PAD_TO, cfg.TEST.PAD_TO))

#         fillHeight = cfg.TEST.PAD_TO - img.shape[0]
#         fillWidth = cfg.TEST.PAD_TO - img.shape[1]

#         padLeft = int(np.round(fillWidth / 2))
#         padRight = int(fillWidth - padLeft)
#         padUp = int(np.round(fillHeight / 2))
#         padDown = int(fillHeight - padUp)

#         img = np.pad(img,
#                     pad_width=((padUp, padDown), (padLeft, padRight)),
#                     mode='reflect')

#         # Normalize keypoint locations
#         kp_norm = []
#         for i, p in enumerate(pts):
#             _p = 2 * np.array([(p[0] + padLeft) / (cfg.TEST.PAD_TO),
#                             (p[1] + padUp) / (cfg.TEST.PAD_TO)]) - 1
#             kp_norm.append(_p)

#         theta = [
#             torch.from_numpy(np.array(kp_norm)).float().squeeze(),
#             torch.from_numpy(scales).float(),
#             torch.from_numpy(np.array([np.deg2rad(o) for o in oris])).float()
#         ]


#         # Extract descriptors
#         imgs, img_keypoints = torch.from_numpy(img).unsqueeze(0).to(self.DEV), \
#             [theta[0].to(self.DEV), theta[1].to(self.DEV), theta[2].to(self.DEV)]

#         # import pdb; pdb.set_trace()   
#         input_filename = 'test'
#         descriptors, patches = self.model({input_filename: imgs}, img_keypoints,
#                                  [input_filename] * len(img_keypoints[0]))

#         descriptors = descriptors.squeeze().detach().cpu().numpy()
#         return cv_kps, descriptors


#     def detect(self, img):
#         return self.detectAndCompute(img)[0]

#     def compute(self, image, keypoints):
#         raise NotImplemented

#     def to(self, device):
#         self.model.to(device)
#         self.DEV = device

#     def match(self, image1, image2):
#         kp0, desc0 = self.detectAndCompute(image1)
#         kp1, desc1 = self.detectAndCompute(image2)
        
#         data = {
#             "descriptors0": desc0.unsqueeze(0),
#             "descriptors1": desc1.unsqueeze(0),
#         }
        
#         response = self.matcher(data)
        
#         m0 = response['matches0'][0]
#         valid = m0 > -1
        
#         mkpts0 = kp0[valid]
#         mkpts1 = kp1[m0[valid]]
        
#         return {
#             'mkpts0': mkpts0,
#             'mkpts1': mkpts1,
#         }
        
#     @property
#     def has_detector(self):
#         return True



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

# References:
#   https://github.com/facebookresearch/dino/blob/master/vision_transformer.py
#   https://github.com/rwightman/pytorch-image-models/tree/master/timm/layers/drop.py


from torch import nn


def drop_path(x, drop_prob: float = 0.0, training: bool = False):
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0:
        random_tensor.div_(keep_prob)
    output = x * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)



import torch
from torch import nn
from torch.distributions import Categorical
# from torch_dimcheck import dimchecked
# from ...submodules.torch_dimcheck import dimchecked

from ..common.structs import Features, NpArray, MatchDistribution
from ..geom import distance_matrix

class ConsistentMatchDistribution(MatchDistribution):
    def __init__(
        self,
        features_1: Features,
        features_2: Features,
        inverse_T: float,
    ):
        self._features_1 = features_1
        self._features_2 = features_2
        self.inverse_T = inverse_T

        distances = distance_matrix(
            self.features_1().desc,
            self.features_2().desc,
        )
        affinity = -inverse_T * distances

        self._cat_I = Categorical(logits=affinity)
        self._cat_T = Categorical(logits=affinity.T)

        self._dense_logp = None
        self._dense_p    = None

    # @dimchecked
    def dense_p(self):
        if self._dense_p is None:
            self._dense_p = self._cat_I.probs * self._cat_T.probs.T

        return self._dense_p

    # @dimchecked
    def dense_logp(self):
        if self._dense_logp is None:
            self._dense_logp = self._cat_I.logits + self._cat_T.logits.T

        return self._dense_logp

    # @dimchecked
    def _select_cycle_consistent(self, left, right):
        indexes = torch.arange(left.shape[0], device=left.device)
        cycle_consistent = right[left] == indexes

        paired_left = left[cycle_consistent]

        return torch.stack([
            right[paired_left],
            paired_left,
        ], dim=0)

    # @dimchecked
    def sample(self):
        samples_I = self._cat_I.sample()
        samples_T = self._cat_T.sample()

        return self._select_cycle_consistent(samples_I, samples_T)

    # @dimchecked
    def mle(self):
        maxes_I = self._cat_I.logits.argmax(dim=1)
        maxes_T = self._cat_T.logits.argmax(dim=1)

        # FIXME UPSTREAM: this detachment is necessary until the bug is fixed
        maxes_I = maxes_I.detach()
        maxes_T = maxes_T.detach()

        return self._select_cycle_consistent(maxes_I, maxes_T)

    def features_1(self) -> Features:
        return self._features_1

    def features_2(self) -> Features:
        return self._features_2

class ConsistentMatcher(torch.nn.Module):
    def __init__(self, inverse_T=1.):
        super(ConsistentMatcher, self).__init__()
        self.inverse_T = nn.Parameter(torch.tensor(inverse_T, dtype=torch.float32))

    def extra_repr(self):
        return f'inverse_T={self.inverse_T.item()}'

    def match_pair(self, features_1: Features, features_2: Features):
        return ConsistentMatchDistribution(features_1, features_2, self.inverse_T)



import torch

# from torch_dimcheck import dimchecked
# from ...submodules.torch_dimcheck import dimchecked

SQRT_2 = 1.414213

# @dimchecked
def distance_matrix(fs1, fs2):
    '''
    Assumes fs1 and fs2 are normalized!
    '''
    return SQRT_2 * (1. - fs1 @ fs2.T).clamp(min=1e-6).sqrt()



# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use

import pdb

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


""" Different samplers, each specifying how to sample pixels for the AP loss.
"""


class FullSampler(nn.Module):
    """ all pixels are selected
        - feats: keypoint descriptors
        - confs: reliability values
    """
    def __init__(self):
        nn.Module.__init__(self)
        self.mode = 'bilinear'
        self.padding = 'zeros'

    @staticmethod
    def _aflow_to_grid(aflow):
        H, W = aflow.shape[2:]
        grid = aflow.permute(0,2,3,1).clone()
        grid[:,:,:,0] *= 2/(W-1)
        grid[:,:,:,1] *= 2/(H-1)
        grid -= 1
        grid[torch.isnan(grid)] = 9e9 # invalids
        return grid
    
    def _warp(self, feats, confs, aflow):
        if isinstance(aflow, tuple): return aflow # result was precomputed
        feat1, feat2 = feats
        conf1, conf2 = confs if confs else (None,None)
    
        B, two, H, W = aflow.shape
        D = feat1.shape[1]
        assert feat1.shape == feat2.shape == (B, D, H, W) # D = 128, B = batch
        assert conf1.shape == conf2.shape == (B, 1, H, W) if confs else True

        # warp img2 to img1
        grid = self._aflow_to_grid(aflow)
        ones2 = feat2.new_ones(feat2[:,0:1].shape)
        feat2to1 = F.grid_sample(feat2, grid, mode=self.mode, padding_mode=self.padding)
        mask2to1 = F.grid_sample(ones2, grid, mode='nearest', padding_mode='zeros')
        conf2to1 = F.grid_sample(conf2, grid, mode=self.mode, padding_mode=self.padding) \
                   if confs else None
        return feat2to1, mask2to1.byte(), conf2to1

    def _warp_positions(self, aflow):
        B, two, H, W = aflow.shape
        assert two == 2
        
        Y = torch.arange(H, device=aflow.device)
        X = torch.arange(W, device=aflow.device)
        XY = torch.stack(torch.meshgrid(Y,X, indexing="ij")[::-1], dim=0)
        XY = XY[None].expand(B, 2, H, W).float()
        
        grid = self._aflow_to_grid(aflow)
        XY2 = F.grid_sample(XY, grid, mode='bilinear', padding_mode='zeros')
        return XY, XY2



class SubSampler (FullSampler):
    """ pixels are selected in an uniformly spaced grid
    """
    def __init__(self, border, subq, subd, perimage=False):
        FullSampler.__init__(self)
        assert subq % subd == 0, 'subq must be multiple of subd'
        self.sub_q = subq
        self.sub_d = subd
        self.border = border
        self.perimage = perimage

    def __repr__(self):
        return "SubSampler(border=%d, subq=%d, subd=%d, perimage=%d)" % (
            self.border, self.sub_q, self.sub_d, self.perimage)

    def __call__(self, feats, confs, aflow):
        feat1, conf1 = feats[0], (confs[0] if confs else None)
        # warp with optical flow in img1 coords
        feat2, mask2, conf2 = self._warp(feats, confs, aflow)
        
        # subsample img1
        slq = slice(self.border, -self.border or None, self.sub_q)
        feat1 = feat1[:, :, slq, slq]
        conf1 = conf1[:, :, slq, slq] if confs else None
        # subsample img2
        sld = slice(self.border, -self.border or None, self.sub_d)
        feat2 = feat2[:, :, sld, sld]
        mask2 = mask2[:, :, sld, sld]
        conf2 = conf2[:, :, sld, sld] if confs else None
        
        B, D, Hq, Wq = feat1.shape
        B, D, Hd, Wd = feat2.shape
        
        # compute gt
        if self.perimage or self.sub_q != self.sub_d:
            # compute ground-truth by comparing pixel indices
            f = feats[0][0:1,0] if self.perimage else feats[0][:,0]
            idxs = torch.arange(f.numel(), dtype=torch.int64, device=feat1.device).view(f.shape)
            idxs1 = idxs[:, slq, slq].reshape(-1,Hq*Wq)
            idxs2 = idxs[:, sld, sld].reshape(-1,Hd*Wd)
            if self.perimage:
                gt = (idxs1[0].view(-1,1) == idxs2[0].view(1,-1))
                gt = gt[None,:,:].expand(B, Hq*Wq, Hd*Wd)
            else :
                gt = (idxs1.view(-1,1) == idxs2.view(1,-1)) 
        else:
            gt = torch.eye(feat1[:,0].numel(), dtype=torch.uint8, device=feat1.device) # always binary for AP loss
        
        # compute all images together
        queries  =  feat1.reshape(B,D,-1) # B x D x (Hq x Wq)
        database =  feat2.reshape(B,D,-1) # B x D x (Hd x Wd)
        if self.perimage:
            queries  =  queries.transpose(1,2) # B x (Hd x Wd) x D
            scores = torch.bmm(queries, database) # B x (Hq x Wq) x (Hd x Wd)
        else:
            queries  =  queries .transpose(1,2).reshape(-1,D) # (B x Hq x Wq) x D
            database =  database.transpose(1,0).reshape(D,-1) # D x (B x Hd x Wd)
            scores = torch.matmul(queries, database) # (B x Hq x Wq) x (B x Hd x Wd)

        # compute reliability
        qconf = (conf1 + conf2)/2 if confs else None

        assert gt.shape == scores.shape
        return scores, gt, mask2, qconf



class NghSampler (FullSampler):
    """ all pixels in a small neighborhood
    """
    def __init__(self, ngh, subq=1, subd=1, ignore=1, border=None):
        FullSampler.__init__(self)
        assert 0 <= ignore < ngh
        self.ngh = ngh
        self.ignore = ignore
        assert subd <= ngh
        self.sub_q = subq
        self.sub_d = subd
        if border is None: border = ngh
        assert border >= ngh, 'border has to be larger than ngh'
        self.border = border

    def __repr__(self):
        return "NghSampler(ngh=%d, subq=%d, subd=%d, ignore=%d, border=%d)" % (
            self.ngh, self.sub_q, self.sub_d, self.ignore, self.border)

    def trans(self, arr, i, j):
        s = lambda i: slice(self.border+i, i-self.border or None, self.sub_q)
        return arr[:,:,s(j),s(i)]

    def __call__(self, feats, confs, aflow):
        feat1, conf1 = feats[0], (confs[0] if confs else None)
        # warp with optical flow in img1 coords
        feat2, mask2, conf2 = self._warp(feats, confs, aflow)
        
        qfeat = self.trans(feat1,0,0)
        qconf = (self.trans(conf1,0,0) + self.trans(conf2,0,0)) / 2 if confs else None
        mask2 = self.trans(mask2,0,0)
        scores_at = lambda i,j: (qfeat * self.trans(feat2,i,j)).sum(dim=1)
        
        # compute scores for all neighbors
        B, D = feat1.shape[:2]
        min_d = self.ignore**2
        max_d = self.ngh**2
        rad = (self.ngh//self.sub_d) * self.ngh # make an integer multiple
        negs = []
        offsets = []
        for j in range(-rad, rad+1, self.sub_d):
          for i in range(-rad, rad+1, self.sub_d):
            if not(min_d < i*i + j*j <= max_d): 
                continue # out of scope
            offsets.append((i,j)) # Note: this list is just for debug
            negs.append( scores_at(i,j) )
        
        scores = torch.stack([scores_at(0,0)] + negs, dim=-1)
        gt = scores.new_zeros(scores.shape, dtype=torch.uint8)
        gt[..., 0] = 1 # only the center point is positive

        return scores, gt, mask2, qconf



class FarNearSampler (FullSampler):
    """ Sample pixels from *both* a small neighborhood *and* far-away pixels.
        
    How it works?
        1) Queries are sampled from img1,
            - at least `border` pixels from borders and 
            - on a grid with step = `subq`
            
        2) Close database pixels 
            - from the corresponding image (img2),
            - within a `ngh` distance radius 
            - on a grid with step = `subd_ngh`
            - ignored if distance to query is >0 and <=`ignore`
            
        3) Far-away database pixels from ,
            - from all batch images in `img2`
            - at least `border` pixels from borders
            - on a grid with step = `subd_far`
    """
    def __init__(self, subq, ngh, subd_ngh, subd_far, border=None, ignore=1, 
                       maxpool_ngh=False ):
        FullSampler.__init__(self)
        border = border or ngh
        assert ignore < ngh < subd_far, 'neighborhood needs to be smaller than far step'
        self.close_sampler = NghSampler(ngh=ngh, subq=subq, subd=subd_ngh, 
                ignore=not(maxpool_ngh), border=border)
        self.faraway_sampler = SubSampler(border=border, subq=subq, subd=subd_far)
        self.maxpool_ngh = maxpool_ngh

    def __repr__(self):
        c,f = self.close_sampler, self.faraway_sampler
        res = "FarNearSampler(subq=%d, ngh=%d" % (c.sub_q, c.ngh)
        res += ", subd_ngh=%d, subd_far=%d" % (c.sub_d, f.sub_d)
        res += ", border=%d, ign=%d" % (f.border, c.ignore)
        res += ", maxpool_ngh=%d" % self.maxpool_ngh
        return res+')'

    def __call__(self, feats, confs, aflow):
        # warp with optical flow in img1 coords
        aflow = self._warp(feats, confs, aflow)

        # sample ngh pixels
        scores1, gt1, msk1, conf1 = self.close_sampler(feats, confs, aflow)
        scores1, gt1 = scores1.view(-1,scores1.shape[-1]), gt1.view(-1,gt1.shape[-1])
        if self.maxpool_ngh:
            # we consider all scores from ngh as potential positives
            scores1, self._cached_maxpool_ngh = scores1.max(dim=1,keepdim=True)
            gt1 = gt1[:, 0:1]

        # sample far pixels
        scores2, gt2, msk2, conf2 = self.faraway_sampler(feats, confs, aflow)
        # assert (msk1 == msk2).all()
        # assert (conf1 == conf2).all()

        return (torch.cat((scores1,scores2),dim=1), 
                torch.cat((gt1,    gt2),    dim=1), 
                msk1, conf1 if confs else None)


class NghSampler2 (nn.Module):
    """ Similar to NghSampler, but doesnt warp the 2nd image.
    Distance to GT =>  0 ... pos_d ... neg_d ... ngh
    Pixel label    =>  + + + + + + 0 0 - - - - - - -
    
    Subsample on query side: if > 0, regular grid
                                < 0, random points 
    In both cases, the number of query points is = W*H/subq**2
    """
    def __init__(self, ngh, subq=1, subd=1, pos_d=0, neg_d=2, border=None,
                       maxpool_pos=True, subd_neg=0):
        nn.Module.__init__(self)
        assert 0 <= pos_d < neg_d <= (ngh if ngh else 99)
        self.ngh = ngh
        self.pos_d = pos_d
        self.neg_d = neg_d
        assert subd <= ngh or ngh == 0
        assert subq != 0
        self.sub_q = subq
        self.sub_d = subd
        self.sub_d_neg = subd_neg
        if border is None: border = ngh
        assert border >= ngh, 'border has to be larger than ngh'
        self.border = border
        self.maxpool_pos = maxpool_pos
        self.precompute_offsets()

    def precompute_offsets(self):
        pos_d2 = self.pos_d**2
        neg_d2 = self.neg_d**2
        rad2 = self.ngh**2
        rad = (self.ngh//self.sub_d) * self.ngh # make an integer multiple
        pos = []
        neg = []
        for j in range(-rad, rad+1, self.sub_d):
          for i in range(-rad, rad+1, self.sub_d):
            d2 = i*i + j*j
            if d2 <= pos_d2:
                pos.append( (i,j) )
            elif neg_d2 <= d2 <= rad2: 
                neg.append( (i,j) )

        self.register_buffer('pos_offsets', torch.LongTensor(pos).view(-1,2).t())
        self.register_buffer('neg_offsets', torch.LongTensor(neg).view(-1,2).t())

    def gen_grid(self, step, aflow):
        B, two, H, W = aflow.shape
        dev = aflow.device
        b1 = torch.arange(B, device=dev)
        if step > 0:
            # regular grid
            x1 = torch.arange(self.border, W-self.border, step, device=dev)
            y1 = torch.arange(self.border, H-self.border, step, device=dev)
            H1, W1 = len(y1), len(x1)
            x1 = x1[None,None,:].expand(B,H1,W1).reshape(-1)
            y1 = y1[None,:,None].expand(B,H1,W1).reshape(-1)
            b1 = b1[:,None,None].expand(B,H1,W1).reshape(-1)
            shape = (B, H1, W1)
        else:
            # randomly spread
            n = (H - 2*self.border) * (W - 2*self.border) // step**2
            x1 = torch.randint(self.border, W-self.border, (n,), device=dev)
            y1 = torch.randint(self.border, H-self.border, (n,), device=dev)
            x1 = x1[None,:].expand(B,n).reshape(-1)
            y1 = y1[None,:].expand(B,n).reshape(-1)
            b1 = b1[:,None].expand(B,n).reshape(-1)
            shape = (B, n)
        return b1, y1, x1, shape

    def forward(self, feats, confs, aflow, **kw):
        B, two, H, W = aflow.shape
        assert two == 2
        feat1, conf1 = feats[0], (confs[0] if confs else None)
        feat2, conf2 = feats[1], (confs[1] if confs else None)
        
        # positions in the first image
        b1, y1, x1, shape = self.gen_grid(self.sub_q, aflow)

        # sample features from first image
        feat1 = feat1[b1, :, y1, x1]
        qconf = conf1[b1, :, y1, x1].view(shape) if confs else None
        
        #sample GT from second image
        b2 = b1
        xy2 = (aflow[b1, :, y1, x1] + 0.5).long().t()
        mask = (0 <= xy2[0]) * (0 <= xy2[1]) * (xy2[0] < W) * (xy2[1] < H)
        mask = mask.view(shape)
        
        def clamp(xy):
            torch.clamp(xy[0], 0, W-1, out=xy[0])
            torch.clamp(xy[1], 0, H-1, out=xy[1])
            return xy
        
        # compute positive scores
        xy2p = clamp(xy2[:,None,:] + self.pos_offsets[:,:,None])
        pscores = (feat1[None,:,:] * feat2[b2, :, xy2p[1], xy2p[0]]).sum(dim=-1).t()
#        xy1p = clamp(torch.stack((x1,y1))[:,None,:] + self.pos_offsets[:,:,None])
#        grid = FullSampler._aflow_to_grid(aflow)
#        feat2p = F.grid_sample(feat2, grid, mode='bilinear', padding_mode='border')
#        pscores = (feat1[None,:,:] * feat2p[b1,:,xy1p[1], xy1p[0]]).sum(dim=-1).t()
        if self.maxpool_pos:
            pscores, pos = pscores.max(dim=1, keepdim=True)
            if confs: 
                sel = clamp(xy2 + self.pos_offsets[:,pos.view(-1)])
                qconf = (qconf + conf2[b2, :, sel[1], sel[0]].view(shape))/2
        
        # compute negative scores
        xy2n = clamp(xy2[:,None,:] + self.neg_offsets[:,:,None])
        nscores = (feat1[None,:,:] * feat2[b2, :, xy2n[1], xy2n[0]]).sum(dim=-1).t()

        if self.sub_d_neg:
            # add distractors from a grid
            b3, y3, x3, _ = self.gen_grid(self.sub_d_neg, aflow)
            distractors = feat2[b3, :, y3, x3]
            dscores = torch.matmul(feat1, distractors.t())
            del distractors
            
            # remove scores that corresponds to positives or nulls
            dis2 = (x3 - xy2[0][:,None])**2 + (y3 - xy2[1][:,None])**2
            dis2 += (b3 != b2[:,None]).long() * self.neg_d**2
            dscores[dis2 < self.neg_d**2] = 0
            
            scores = torch.cat((pscores, nscores, dscores), dim=1)
        else:
            # concat everything
            scores = torch.cat((pscores, nscores), dim=1)

        gt = scores.new_zeros(scores.shape, dtype=torch.uint8)
        gt[:, :pscores.shape[1]] = 1

        return scores, gt, mask, qconf











import argparse
import random
import os
import numpy as np
import torch
from omegaconf import OmegaConf
from easy_local_features.matching.nearest_neighbor import NearestNeighborMatcher
from easy_local_features.feature.basemodel import BaseExtractor, MethodType
from easy_local_features.utils import ops, download
from easy_local_features.submodules.git_rekd.model import (
    load_detector,
    apply_homography_to_points,
    get_point_coordinates,
    remove_borders,
    apply_nms,
    upsample_pyramid,
)
import warnings


def fix_randseed(randseed):
    r"""Fix random seed"""
    random.seed(randseed)
    np.random.seed(randseed)
    torch.manual_seed(randseed)
    torch.cuda.manual_seed(randseed)
    torch.cuda.manual_seed_all(randseed)
    torch.backends.cudnn.benchmark, torch.backends.cudnn.deterministic = False, True
    # torch.backends.cudnn.benchmark, torch.backends.cudnn.deterministic = True, False


def get_config(jupyter=False):
    parser = argparse.ArgumentParser(description="Train REKD Architecture")

    ## basic configuration
    parser.add_argument(
        "--data_dir",
        type=str,
        default="../ImageNet2012/ILSVRC2012_img_val",  # default='path-to-ImageNet',
        help="The root path to the data from which the synthetic dataset will be created.",
    )
    parser.add_argument(
        "--synth_dir",
        type=str,
        default="",
        help="The path to save the generated sythetic image pairs.",
    )
    parser.add_argument(
        "--log_dir",
        type=str,
        default="trained_models/weights",
        help="The path to save the REKD weights.",
    )
    parser.add_argument(
        "--load_dir",
        type=str,
        default="",
        help="Set saved model parameters if resume training is desired.",
    )
    parser.add_argument(
        "--exp_name",
        type=str,
        default="REKD",
        help="The Rotaton-equivaraiant Keypoint Detection (REKD) experiment name",
    )
    ## network architecture
    parser.add_argument(
        "--factor_scaling_pyramid",
        type=float,
        default=1.2,
        help="The scale factor between the multi-scale pyramid levels in the architecture.",
    )
    parser.add_argument(
        "--group_size",
        type=int,
        default=36,
        help="The number of groups for the group convolution.",
    )
    parser.add_argument(
        "--dim_first",
        type=int,
        default=2,
        help="The number of channels of the first layer",
    )
    parser.add_argument(
        "--dim_second",
        type=int,
        default=2,
        help="The number of channels of the second layer",
    )
    parser.add_argument(
        "--dim_third",
        type=int,
        default=2,
        help="The number of channels of the thrid layer",
    )
    ## network training
    parser.add_argument("--batch_size", type=int, default=16, help="The batch size for training.")
    parser.add_argument("--num_epochs", type=int, default=20, help="Number of epochs for training.")
    ## Loss function
    parser.add_argument(
        "--init_initial_learning_rate",
        type=float,
        default=1e-3,
        help="The init initial learning rate value.",
    )
    parser.add_argument("--MSIP_sizes", type=str, default="8,16,24,32,40", help="MSIP sizes.")
    parser.add_argument(
        "--MSIP_factor_loss",
        type=str,
        default="256.0,64.0,16.0,4.0,1.0",
        help="MSIP loss balancing parameters.",
    )
    parser.add_argument("--ori_loss_balance", type=float, default=100.0, help="")
    ## Dataset generation
    parser.add_argument(
        "--patch_size",
        type=int,
        default=192,
        help="The patch size of the generated dataset.",
    )
    parser.add_argument(
        "--max_angle",
        type=int,
        default=180,
        help="The max angle value for generating a synthetic view to train REKD.",
    )
    parser.add_argument(
        "--min_scale",
        type=float,
        default=1.0,
        help="The min scale value for generating a synthetic view to train REKD.",
    )
    parser.add_argument(
        "--max_scale",
        type=float,
        default=1.0,
        help="The max scale value for generating a synthetic view to train REKD.",
    )
    parser.add_argument(
        "--max_shearing",
        type=float,
        default=0.0,
        help="The max shearing value for generating a synthetic view to train REKD.",
    )
    parser.add_argument(
        "--num_training_data",
        type=int,
        default=9000,
        help="The number of the generated dataset.",
    )
    parser.add_argument(
        "--is_debugging",
        type=bool,
        default=False,
        help="Set variable to True if you desire to train network on a smaller dataset.",
    )
    ## For eval/inference
    parser.add_argument(
        "--num_points",
        type=int,
        default=1500,
        help="the number of points at evaluation time.",
    )
    parser.add_argument("--pyramid_levels", type=int, default=5, help="downsampling pyramid levels.")
    parser.add_argument("--upsampled_levels", type=int, default=2, help="upsampling image levels.")
    parser.add_argument(
        "--nms_size",
        type=int,
        default=15,
        help="The NMS size for computing the validation repeatability.",
    )
    parser.add_argument(
        "--border_size",
        type=int,
        default=15,
        help="The number of pixels to remove from the borders to compute the repeatability.",
    )
    ## For HPatches evaluation
    parser.add_argument(
        "--hpatches_path",
        type=str,
        default="./datasets/hpatches-sequences-release",
        help="dataset ",
    )
    parser.add_argument(
        "--eval_split",
        type=str,
        default="debug",
        help="debug, view, illum, full, debug_view, debug_illum ...",
    )
    parser.add_argument("--descriptor", type=str, default="hardnet", help="hardnet, sosnet, hynet")

    args, weird_args = parser.parse_known_args() if not jupyter else parser.parse_args(args=[])

    fix_randseed(12345)

    if args.synth_dir == "":
        args.synth_dir = "datasets/synth_data"

    args.MSIP_sizes = [int(i) for i in args.MSIP_sizes.split(",")]
    args.MSIP_factor_loss = [float(i) for i in args.MSIP_factor_loss.split(",")]

    return args

WEIGHTS = "https://github.com/felipecadar/easy-local-features-baselines/releases/download/redk/best_model.pt"
class REKD_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE
    default_conf = {
        "num_keypoints": 1500,
        "pyramid_levels": 5,
        "upsampled_levels": 2,
        "border_size": 15,
        "nms_size": 15,
        "weights": None,
        "resize": None,
    }

    def __init__(self, conf={}):
        super().__init__()
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)

        # configurations
        self.default_num_points = conf.num_keypoints
        self.pyramid_levels = conf.pyramid_levels
        self.upsampled_levels = conf.upsampled_levels
        self.resize = conf.resize
        self.border_size = conf.border_size
        self.nms_size = conf.nms_size
        self.desc_scale_factor = 2.0
        self.scale_factor_levels = np.sqrt(2)

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Create args object for model loading
        args = get_config()
        args.weights = conf.weights
        if args.weights is None or (not os.path.isfile(args.weights)):
            args.weights = str(download.downloadModel("redk", "release_group36_f2_s2_t2.log/best_model.pt", WEIGHTS))
            print(f"Using default REKD weights from {args.weights}")

        args.resize = conf.resize

        self.model = load_detector(args, self.device)
        self.levels = self.pyramid_levels + self.upsampled_levels + 1

        self.matcher = NearestNeighborMatcher()

    def to_normalized_coords(self, keypoints, H, W):
        """Convert pixel coordinates to normalized coordinates [-1, 1]"""
        kpts = keypoints.clone()
        kpts[..., 0] = 2 * kpts[..., 0] / (W - 1) - 1  # x
        kpts[..., 1] = 2 * kpts[..., 1] / (H - 1) - 1  # y
        return kpts

    def from_normalized_coords(self, keypoints, H, W):
        """Convert normalized coordinates [-1, 1] to pixel coordinates"""
        kpts = keypoints.clone()
        kpts[..., 0] = (kpts[..., 0] + 1) * (W - 1) / 2  # x
        kpts[..., 1] = (kpts[..., 1] + 1) * (H - 1) / 2  # y
        return kpts

    def detectAndCompute(self, image, return_dict=False):
        return NotImplementedError

    def detect(self, image):
        """Detect keypoints only"""
        image = ops.prepareImage(image).to(self.device)
        img_np = image[0, 0].cpu().numpy() if len(image.shape) == 4 else image[0].cpu().numpy()
        kps = self._detect_keypoints(img_np)
        return kps

    def compute(self, image, keypoints):
        return NotImplementedError

    def _detect_keypoints(self, image):
        """Internal method to detect keypoints"""
        # Ensure image has the correct shape (C, H, W)
        if len(image.shape) == 2:
            # Add channel dimension for grayscale
            image = image[np.newaxis, ...]

        one, H, W = image.shape
        score_maps, ori_maps = self._compute_score_maps(image)
        im_pts = self._estimate_keypoint_coordinates(score_maps, num_points=self.default_num_points)
        pixel_coords = im_pts[..., :2]
        pixel_coords = torch.tensor(pixel_coords).to(self.device)
        return pixel_coords

    def to(self, device):
        """Move model to device"""
        self.model = self.model.to(device)
        self.device = device
        return self

    @property
    def has_detector(self):
        """REKD has a detector"""
        return True

    def _compute_score_maps(self, image):
        from skimage.transform import pyramid_gaussian

        pyramid = pyramid_gaussian(image, max_layer=self.pyramid_levels, downscale=self.scale_factor_levels)
        up_pyramid = upsample_pyramid(
            image,
            upsampled_levels=self.upsampled_levels,
            scale_factor_levels=self.scale_factor_levels,
        )

        score_maps = {}
        ori_maps = {}
        for j, down_image in enumerate(pyramid):  ## Pyramid is downsampling images.
            key_idx = j + 1 + self.upsampled_levels
            score_maps, ori_maps = self._obtain_feature_maps(down_image, key_idx, score_maps, ori_maps)

        if self.upsampled_levels:
            for j, up_image in enumerate(up_pyramid):  ## Upsample levels is for upsampling images.
                key_idx = j + 1
                score_maps, ori_maps = self._obtain_feature_maps(up_image, key_idx, score_maps, ori_maps)

        return score_maps, ori_maps

    def _obtain_feature_maps(self, im, key_idx, score_maps, ori_maps):
        im = torch.tensor(im).unsqueeze(0).to(torch.float32).to(self.device)
        im_scores, ori_map = self.model(im)
        im_scores = remove_borders(im_scores[0, 0, :, :].cpu().detach().numpy(), borders=self.border_size)

        score_maps["map_" + str(key_idx)] = im_scores
        ori_maps["map_" + str(key_idx)] = ori_map

        return score_maps, ori_maps

    def _estimate_keypoint_coordinates(self, score_maps, num_points=None):
        num_points = num_points if num_points is not None else self.default_num_points
        point_level = []
        tmp = 0.0
        factor_points = self.scale_factor_levels**2
        for idx_level in range(self.levels):
            tmp += factor_points ** (-1 * (idx_level - self.upsampled_levels))
            point_level.append(self.default_num_points * factor_points ** (-1 * (idx_level - self.upsampled_levels)))

        point_level = np.asarray(list(map(lambda x: int(x / tmp) + 1, point_level)))

        im_pts = []
        for idx_level in range(self.levels):
            scale_value = self.scale_factor_levels ** (idx_level - self.upsampled_levels)
            scale_factor = 1.0 / scale_value

            h_scale = np.asarray([[scale_factor, 0.0, 0.0], [0.0, scale_factor, 0.0], [0.0, 0.0, 1.0]])
            h_scale_inv = np.linalg.inv(h_scale)
            h_scale_inv = h_scale_inv / h_scale_inv[2, 2]

            num_points_level = point_level[idx_level]
            if idx_level > 0:
                res_points = int(np.asarray([point_level[a] for a in range(0, idx_level + 1)]).sum() - len(im_pts))
                num_points_level = res_points

            ## to make the output score map derive more keypoints
            score_map = score_maps["map_" + str(idx_level + 1)]

            im_scores = apply_nms(score_map, self.nms_size)
            im_pts_tmp = get_point_coordinates(im_scores, num_points=num_points_level)
            im_pts_tmp = apply_homography_to_points(im_pts_tmp, h_scale_inv)

            if not idx_level:
                im_pts = im_pts_tmp
            else:
                im_pts = np.concatenate((im_pts, im_pts_tmp), axis=0)

        im_pts = im_pts[(-1 * im_pts[:, 3]).argsort()]
        im_pts = im_pts[:num_points]

        return im_pts

if __name__ == "__main__":
    from easy_local_features.utils import io, vis, ops

    detector = REKD_baseline({"num_keypoints": 512})

    img0 = io.fromPath("tests/assets/megadepth0.jpg")
    img1 = io.fromPath("tests/assets/megadepth1.jpg")
    
    img0, _ = ops.resize_short_edge(img0, 480)
    img1, _ = ops.resize_short_edge(img1, 480)

    kps0 = detector.detect(img0)
    kps1 = detector.detect(img1)

    vis.plot_pair(img0, img1)
    vis.plot_keypoints(keypoints0=kps0)
    vis.plot_keypoints(keypoints1=kps1)
    vis.show()



'''=================================================
@Project -> File   r2d2 -> extract_localization
@IDE    PyCharm
@Author fx221@cam.ac.uk
@Date   13/07/2022 09:59
=================================================='''

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)

def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)

class ResBlock(nn.Module):
    def __init__(self, inplanes, outplanes, stride=1, groups=32, dilation=1, norm_layer=None):
        super(ResBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self.conv1 = conv1x1(inplanes, outplanes)
        self.bn1 = norm_layer(outplanes)
        self.conv2 = conv3x3(outplanes, outplanes, stride, groups, dilation)
        self.bn2 = norm_layer(outplanes)
        self.conv3 = conv1x1(outplanes, outplanes)
        self.bn3 = norm_layer(outplanes)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out += identity
        out = self.relu(out)

        return out

def batch_normalization(channels, relu=False):
    if relu:
        return nn.Sequential(
            nn.BatchNorm2d(channels, affine=False, track_running_stats=True),
            nn.ReLU(), )
    else:
        return nn.Sequential(
            nn.BatchNorm2d(channels, affine=False, track_running_stats=True), )

def conv(in_channels, out_channels, kernel_size=3, stride=1, padding=1, relu=True, use_bn=True, dilation=1):
    if not use_bn:
        if relu:
            return nn.Sequential(
                nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                          padding=padding, dilation=dilation),
                nn.ReLU(),
            )
        else:
            return nn.Sequential(
                nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                          padding=padding, dilation=dilation)
            )
    else:
        if relu:
            return nn.Sequential(
                nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                          padding=padding, dilation=dilation),
                nn.BatchNorm2d(out_channels, affine=False),
                nn.ReLU(),
            )
        else:
            return nn.Sequential(
                nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                          padding=padding, dilation=dilation),
                nn.BatchNorm2d(out_channels, affine=False),
                # nn.ReLU(),
            )

class ResSegNet(nn.Module):
    def __init__(self, outdim=128, require_feature=False, require_stability=False, ms_detector=True):
        super().__init__()
        self.outdim = outdim
        self.require_feature = require_feature
        self.require_stability = require_stability
        self.ms_detector = ms_detector

        d1, d2, d3, d4, d5, d6 = 64, 128, 256, 256, 256, 256
        self.conv1a = conv(in_channels=3, out_channels=d1, kernel_size=3, relu=True, use_bn=True)
        self.conv1b = conv(in_channels=d1, out_channels=d1, kernel_size=3, stride=2, relu=False, use_bn=False)
        self.bn1b = batch_normalization(channels=d1, relu=True)

        self.conv2a = conv(in_channels=d1, out_channels=d2, kernel_size=3, relu=True, use_bn=True)
        self.conv2b = conv(in_channels=d2, out_channels=d2, kernel_size=3, stride=2, relu=False, use_bn=False)
        self.bn2b = batch_normalization(channels=d2, relu=True)

        self.conv3a = conv(in_channels=d2, out_channels=d3, kernel_size=3, relu=True, use_bn=True)
        self.conv3b = conv(in_channels=d3, out_channels=d3, kernel_size=3, relu=False, use_bn=False)
        self.bn3b = batch_normalization(channels=d3, relu=True)

        self.conv4 = nn.Sequential(
            ResBlock(inplanes=256, outplanes=256, groups=32),
            ResBlock(inplanes=256, outplanes=256, groups=32),
            ResBlock(inplanes=256, outplanes=256, groups=32),
        )

        self.convPa = nn.Sequential(
            torch.nn.Conv2d(256, 256, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            torch.nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1),
        )
        self.convDa = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1)
        )

        self.convPb = torch.nn.Conv2d(256, 65, kernel_size=1, stride=1, padding=0)
        self.convDb = torch.nn.Conv2d(256, outdim, kernel_size=1, stride=1, padding=0)

        if self.require_stability:
            self.ConvSta = nn.Conv2d(256, 1, kernel_size=1)

    def det(self, x):
        out1a = self.conv1a(x)
        out1b = self.conv1b(out1a)
        out1c = self.bn1b(out1b)

        out2a = self.conv2a(out1c)
        out2b = self.conv2b(out2a)
        out2c = self.bn2b(out2b)

        out3a = self.conv3a(out2c)
        out3b = self.conv3b(out3a)
        out3c = self.bn3b(out3b)

        out4 = self.conv4(out3c)

        cPa = self.convPa(out4)
        semi = self.convPb(cPa)
        semi = torch.exp(semi)
        semi_norm = semi / (torch.sum(semi, dim=1, keepdim=True) + .00001)
        score = semi_norm[:, :-1, :, :]
        Hc, Wc = score.size(2), score.size(3)
        score = score.permute([0, 2, 3, 1])
        score = score.view(score.size(0), Hc, Wc, 8, 8)
        score = score.permute([0, 1, 3, 2, 4])
        score = score.contiguous().view(score.size(0), 1, Hc * 8, Wc * 8)

        # Descriptor Head
        cDa = self.convDa(out4)
        desc = self.convDb(cDa)
        desc = F.normalize(desc, dim=1)

        if self.require_stability:
            # out2b_up = F.interpolate(out2b, size=(x.shape[2], x.shape[3]), mode='bilinear', align_corners=False)
            # out4c_up = F.interpolate(out4c, size=(x.shape[2], x.shape[3]), mode='bilinear', align_corners=False)
            # stability = torch.sigmoid(self.ConvSta(torch.cat([out1b, out2b_up, out4c_up], dim=1)))
            stability = torch.sigmoid(self.ConvSta(out4))
            stability = F.interpolate(stability, size=(x.shape[2], x.shape[3]), mode='bilinear')
        else:
            stability = None

        if self.require_feature and self.training:
            return score, stability, desc, (out1c, out2c, out3c)
        else:
            return score, stability, desc

    def det_train(self, x):
        out1a = self.conv1a(x)
        out1b = self.conv1b(out1a)
        out1c = self.bn1b(out1b)

        out2a = self.conv2a(out1c)
        out2b = self.conv2b(out2a)
        out2c = self.bn2b(out2b)

        out3a = self.conv3a(out2c)
        out3b = self.conv3b(out3a)
        out3c = self.bn3b(out3b)

        out4 = self.conv4(out3c)

        cPa = self.convPa(out4)
        semi = self.convPb(cPa)
        semi = torch.exp(semi)
        semi_norm = semi / (torch.sum(semi, dim=1, keepdim=True) + .00001)
        score = semi_norm[:, :-1, :, :]
        Hc, Wc = score.size(2), score.size(3)
        score = score.permute([0, 2, 3, 1])
        score = score.view(score.size(0), Hc, Wc, 8, 8)
        score = score.permute([0, 1, 3, 2, 4])
        score = score.contiguous().view(score.size(0), 1, Hc * 8, Wc * 8)

        # Descriptor Head
        cDa = self.convDa(out4)
        desc = self.convDb(cDa)
        desc = F.normalize(desc, dim=1)

        if self.require_stability:
            # out2b_up = F.interpolate(out2b, size=(x.shape[2], x.shape[3]), mode='bilinear', align_corners=False)
            # out4c_up = F.interpolate(out4c, size=(x.shape[2], x.shape[3]), mode='bilinear', align_corners=False)
            # stability = torch.sigmoid(self.ConvSta(torch.cat([out1b, out2b_up, out4c_up], dim=1)))
            stability = torch.sigmoid(self.ConvSta(out4))
            stability = F.interpolate(stability, size=(x.shape[2], x.shape[3]), mode='bilinear')
        else:
            stability = None

        if self.require_feature and self.training:
            return score, semi_norm, stability, desc, (out2c, out3c)
        else:
            return score, semi_norm, stability, desc

    def forward(self, batch):
        # b = batch['image1'].size(0)
        x = torch.cat([batch['image1'], batch['image2']], dim=0)

        if self.require_feature:
            score, semi, stability, desc, seg_feats = self.det_train(x)
            return {
                "reliability": score,
                "score": score,
                "semi": semi,
                "stability": stability,
                "desc": desc,
                "pred_feats": seg_feats,
            }
        else:
            score, semi, stability, desc = self.det_train(x)
            return {
                "reliability": score,
                "score": score,
                "semi": semi,
                "stability": stability,
                "desc": desc,
            }

class ResSegNetV2(nn.Module):
    def __init__(self, outdim=128, require_feature=False, require_stability=False, ms_detector=True):
        super().__init__()
        self.outdim = outdim
        self.require_feature = require_feature
        self.require_stability = require_stability
        self.ms_detector = ms_detector

        d1, d2, d3, d4, d5, d6 = 64, 128, 256, 256, 256, 256
        self.conv1a = conv(in_channels=3, out_channels=d1, kernel_size=3, relu=True, use_bn=True)
        self.conv1b = conv(in_channels=d1, out_channels=d1, kernel_size=3, stride=2, relu=False, use_bn=False)
        self.bn1b = batch_normalization(channels=d1, relu=True)

        self.conv2a = conv(in_channels=d1, out_channels=d2, kernel_size=3, relu=True, use_bn=True)
        self.conv2b = conv(in_channels=d2, out_channels=d2, kernel_size=3, stride=2, relu=False, use_bn=False)
        self.bn2b = batch_normalization(channels=d2, relu=True)

        self.conv3a = conv(in_channels=d2, out_channels=d3, kernel_size=3, relu=True, use_bn=True)
        self.conv3b = conv(in_channels=d3, out_channels=d3, kernel_size=3, relu=False, use_bn=False)
        self.bn3b = batch_normalization(channels=d3, relu=True)

        self.conv4 = nn.Sequential(
            ResBlock(inplanes=256, outplanes=256, groups=32),
            ResBlock(inplanes=256, outplanes=256, groups=32),
            ResBlock(inplanes=256, outplanes=256, groups=32),
        )

        self.convPa = nn.Sequential(
            torch.nn.Conv2d(256, 256, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            torch.nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1),
        )
        self.convDa = nn.Sequential(
            nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1)
        )

        self.convPb = torch.nn.Conv2d(256, 65, kernel_size=1, stride=1, padding=0)
        self.convDb = torch.nn.Conv2d(256, outdim, kernel_size=1, stride=1, padding=0)

        if self.require_stability:
            self.ConvSta = nn.Conv2d(256, 3, kernel_size=1)

    def cls_to_value(self, x):
        # 0 - 0.1, 1-0.5, 2-1.0
        cls = torch.max(x, dim=1, keepdim=True)[1]
        stab = torch.ones_like(cls).float()
        stab[cls == 0] = 0.1
        stab[cls == 1] = 0.5
        return stab

    def det(self, x):
        out1a = self.conv1a(x)
        out1b = self.conv1b(out1a)
        out1c = self.bn1b(out1b)

        out2a = self.conv2a(out1c)
        out2b = self.conv2b(out2a)
        out2c = self.bn2b(out2b)

        out3a = self.conv3a(out2c)
        out3b = self.conv3b(out3a)
        out3c = self.bn3b(out3b)

        out4 = self.conv4(out3c)

        cPa = self.convPa(out4)
        semi = self.convPb(cPa)
        semi = torch.exp(semi)
        semi_norm = semi / (torch.sum(semi, dim=1, keepdim=True) + .00001)
        score = semi_norm[:, :-1, :, :]
        Hc, Wc = score.size(2), score.size(3)
        score = score.permute([0, 2, 3, 1])
        score = score.view(score.size(0), Hc, Wc, 8, 8)
        score = score.permute([0, 1, 3, 2, 4])
        score = score.contiguous().view(score.size(0), 1, Hc * 8, Wc * 8)

        # Descriptor Head
        cDa = self.convDa(out4)
        desc = self.convDb(cDa)
        desc = F.normalize(desc, dim=1)

        if self.require_stability:
            stability = self.ConvSta(out4)
            stability = F.interpolate(stability, size=(x.shape[2], x.shape[3]), mode='bilinear')
            stability = self.cls_to_value(x=stability)
        else:
            stability = None

        if self.require_feature and self.training:
            return score, stability, desc, (out1c, out2c, out3c)
        else:
            return score, stability, desc

    def det_train(self, x):
        out1a = self.conv1a(x)
        out1b = self.conv1b(out1a)
        out1c = self.bn1b(out1b)

        out2a = self.conv2a(out1c)
        out2b = self.conv2b(out2a)
        out2c = self.bn2b(out2b)

        out3a = self.conv3a(out2c)
        out3b = self.conv3b(out3a)
        out3c = self.bn3b(out3b)

        out4 = self.conv4(out3c)

        cPa = self.convPa(out4)
        semi = self.convPb(cPa)
        semi = torch.exp(semi)
        semi_norm = semi / (torch.sum(semi, dim=1, keepdim=True) + .00001)
        score = semi_norm[:, :-1, :, :]
        Hc, Wc = score.size(2), score.size(3)
        score = score.permute([0, 2, 3, 1])
        score = score.view(score.size(0), Hc, Wc, 8, 8)
        score = score.permute([0, 1, 3, 2, 4])
        score = score.contiguous().view(score.size(0), 1, Hc * 8, Wc * 8)

        # Descriptor Head
        cDa = self.convDa(out4)
        desc = self.convDb(cDa)
        desc = F.normalize(desc, dim=1)

        if self.require_stability:
            # out2b_up = F.interpolate(out2b, size=(x.shape[2], x.shape[3]), mode='bilinear', align_corners=False)
            # out4c_up = F.interpolate(out4c, size=(x.shape[2], x.shape[3]), mode='bilinear', align_corners=False)
            # stability = torch.sigmoid(self.ConvSta(torch.cat([out1b, out2b_up, out4c_up], dim=1)))
            stability = self.ConvSta(out4)
            stability = F.interpolate(stability, size=(x.shape[2], x.shape[3]), mode='bilinear')
            score = score * self.cls_to_value(x=stability)

            stability = torch.softmax(stability, dim=1)
        else:
            stability = None

        if self.require_feature and self.training:
            return score, semi_norm, stability, desc, (out2c, out3c)
        else:
            return score, semi_norm, stability, desc

    def forward(self, batch):
        x = torch.cat([batch['image1'], batch['image2']], dim=0)

        if self.require_feature:
            score, semi, stability, desc, seg_feats = self.det_train(x)
            return {
                "reliability": score,
                "score": score,
                "semi": semi,
                "stability": stability,
                "desc": desc,
                "pred_feats": seg_feats,
            }
        else:
            score, semi, stability, desc = self.det_train(x)
            return {
                "reliability": score,
                "score": score,
                "semi": semi,
                "stability": stability,
                "desc": desc,
            }



import torch

from easy_local_features.submodules.git_r2d2.extract import NonMaxSuppression, extract_multiscale, load_network
from easy_local_features.submodules.git_r2d2.nets.patchnet import *

# Pretrained models
# -----------------
# For your convenience, we provide five pre-trained models in the `models/` folder:
#  - `r2d2_WAF_N16.pt`: this is the model used in most experiments of the paper (on HPatches `MMA@3=0.686`). It was trained with Web images (`W`), Aachen day-time images (`A`) and Aachen optical flow pairs (`F`)
#  - `r2d2_WASF_N16.pt`: this is the model used in the visual localization experiments (on HPatches `MMA@3=0.721`). It was trained with Web images (`W`), Aachen day-time images (`A`), Aachen day-night synthetic pairs (`S`), and Aachen optical flow pairs (`F`).
#  - `r2d2_WASF_N8_big.pt`: Same than previous model, but trained with `N=8` instead of `N=16` in the repeatability loss. In other words, it outputs a higher density of keypoints. This can be interesting for certain applications like visual localization, but it implies a drop in MMA since keypoints gets slighlty less reliable.
#  - `faster2d2_WASF_N16.pt`: The Fast-R2D2 equivalent of r2d2_WASF_N16.pt
#  - `faster2d2_WASF_N8_big.pt`: The Fast-R2D2 equivalent of r2d2_WASF_N8.pt

# model name	model size  (#weights)	number of keypoints	MMA@3 on HPatches
# r2d2_WAF_N16.pt	0.5M	5K	0.686
# r2d2_WASF_N16.pt	0.5M	5K	0.721
# r2d2_WASF_N8_big.pt	1.0M	10K	0.692
# faster2d2_WASF_N8_big.pt	1.0M	5K	0.650

models_URLs = {
    "faster2d2_WASF_N8_big": "https://github.com/naver/r2d2/raw/master/models/faster2d2_WASF_N16.pt",
    "faster2d2_WASF_N16": "https://github.com/naver/r2d2/raw/master/models/faster2d2_WASF_N8_big.pt",
    "r2d2_WAF_N16": "https://github.com/naver/r2d2/raw/master/models/r2d2_WAF_N16.pt",
    "r2d2_WASF_N8_big": "https://github.com/naver/r2d2/raw/master/models/r2d2_WASF_N16.pt",
    "r2d2_WASF_N16": "https://github.com/naver/r2d2/raw/master/models/r2d2_WASF_N8_big.pt",
}

from omegaconf import OmegaConf

from ..matching.nearest_neighbor import NearestNeighborMatcher
from ..utils import ops
from ..utils.download import downloadModel
from .basemodel import BaseExtractor, MethodType
from typing import TypedDict, Optional


class R2D2Config(TypedDict):
    model_name: str
    top_k: int
    detection_threshold: float
    nms_radius: int
    rel_thr: float
    rep_thr: float
    scale_f: float
    min_scale: float
    max_scale: float
    min_size: int
    max_size: int
    pretrained_weigts: str
    model_path: Optional[str]


class R2D2_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE

    default_conf: R2D2Config = {
        "model_name": "r2d2",
        "top_k": 2048,
        "detection_threshold": 0.2,
        "nms_radius": 4,
        "rel_thr": 0.7,
        "rep_thr": 0.7,
        "scale_f": 2**0.25,
        "min_scale": 0.0,
        "max_scale": 1,
        "min_size": 256,
        "max_size": 1024,
        "pretrained_weigts": "r2d2_WASF_N16",
        "model_path": None,
    }

    def __init__(self, conf: R2D2Config = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)

        self.top_k = conf.top_k
        self.rel_thr = conf.rel_thr
        self.rep_thr = conf.rep_thr
        self.scale_f = conf.scale_f
        self.min_scale = conf.min_scale
        self.max_scale = conf.max_scale
        self.min_size = conf.min_size
        self.max_size = conf.max_size
        pretrained_weigts = conf.pretrained_weigts

        self.DEV = torch.device("cpu")

        url = models_URLs[pretrained_weigts]
        model_path = downloadModel("r2d2", pretrained_weigts, url)

        # load the network...
        self.model = load_network(model_path)

        # create the non-maxima detector
        self.detector = NonMaxSuppression(rel_thr=self.rel_thr, rep_thr=self.rep_thr)

        self.matcher = NearestNeighborMatcher()

    def _toTorch(self, img):
        return img

    def detectAndCompute(self, img, return_dict=False):
        img = ops.prepareImage(img, imagenet=True).to(self.DEV)

        # extract keypoints/descriptors for a single image
        xys, desc, scores = extract_multiscale(
            self.model,
            img,
            self.detector,
            scale_f=self.scale_f,
            min_scale=self.min_scale,
            max_scale=self.max_scale,
            min_size=self.min_size,
            max_size=self.max_size,
            verbose=False,
        )

        # idxs = scores.argsort()[-self.top_k or None:]
        idxs = torch.argsort(scores, descending=True)[: self.top_k]

        keypoints = xys[idxs]
        keypoints, scales = keypoints.split([2, 1], dim=1)
        descriptors = desc[idxs]
        scores = scores[idxs]

        keypoints = keypoints.unsqueeze(0)
        descriptors = descriptors.unsqueeze(0)
        scales = scales.unsqueeze(0)
        scores = scores.unsqueeze(0)

        if return_dict:
            return {
                "keypoints": keypoints,
                "descriptors": descriptors,
                "scores": scores,
                "scales": scales,
            }

        return keypoints, descriptors

    def detect(self, img):
        return self.detectAndCompute(img)[0]

    def compute(self, image, keypoints):
        raise NotImplementedError

    def to(self, device):
        self.model.to(device)
        self.DEV = device

    @property
    def has_detector(self):
        return True



import torch
import numpy as np

# from ...submodules.torch_dimcheck import dimchecked
from ..unets import Unet, thin_setup

from .detector import Detector
from ..common.structs import NpArray, Features

DEFAULT_SETUP = {**thin_setup, 'bias': True, 'padding': True}

class DISK(torch.nn.Module):
    def __init__(
        self,
        desc_dim=128,
        window=8,
        setup=DEFAULT_SETUP,
        kernel_size=5,
    ):
        super(DISK, self).__init__()

        self.desc_dim = desc_dim
        self.unet = Unet(
            in_features=3, size=kernel_size,
            down=[16, 32, 64, 64, 64],
            up=[64, 64, 64, desc_dim+1],
            setup=setup,
        )
        self.detector = Detector(window=window)

    # @dimchecked
    def _split(self, unet_output):
        '''
        Splits the raw Unet output into descriptors and detection heatmap.
        '''
        assert unet_output.shape[1] == self.desc_dim + 1

        descriptors = unet_output[:, :self.desc_dim]
        heatmap     = unet_output[:, self.desc_dim:]

        return descriptors, heatmap

    # @dimchecked
    def features(
        self,
        images,
        kind='rng',
        **kwargs
    ) -> NpArray[Features]:
        ''' allowed values for `kind`:
            * rng
            * nms
        '''

        B = images.shape[0]
        try:
            descriptors, heatmaps = self._split(self.unet(images))
        except RuntimeError as e:
            if 'Trying to downsample' in str(e):
                msg = ('U-Net failed because the input is of wrong shape. With '
                       'a n-step U-Net (n == 4 by default), input images have '
                       'to have height and width as multiples of 2^n (16 by '
                       'default).')
                raise RuntimeError(msg) from e
            else:
                raise

        keypoints = {
            'rng': self.detector.sample,
            'nms': self.detector.nms,
        }[kind](heatmaps, **kwargs)

        features = []
        for i in range(B):
            features.append(keypoints[i].merge_with_descriptors(descriptors[i]))

        return np.array(features, dtype=object)



# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use


import os, pdb
from PIL import Image
import numpy as np
import torch

from .tools import common
from .tools.dataloader import norm_RGB
from .nets.patchnet import *


def load_network(model_fn): 
    checkpoint = torch.load(model_fn)
    net = eval(checkpoint['net'])
    nb_of_weights = common.model_size(net)
    # initialization
    weights = checkpoint['state_dict']
    net.load_state_dict({k.replace('module.',''):v for k,v in weights.items()})
    return net.eval()


class NonMaxSuppression (torch.nn.Module):
    def __init__(self, rel_thr=0.7, rep_thr=0.7):
        nn.Module.__init__(self)
        self.max_filter = torch.nn.MaxPool2d(kernel_size=3, stride=1, padding=1)
        self.rel_thr = rel_thr
        self.rep_thr = rep_thr
    
    def forward(self, reliability, repeatability, **kw):
        assert len(reliability) == len(repeatability) == 1
        reliability, repeatability = reliability[0], repeatability[0]

        # local maxima
        maxima = (repeatability == self.max_filter(repeatability))

        # remove low peaks
        maxima *= (repeatability >= self.rep_thr)
        maxima *= (reliability   >= self.rel_thr)

        return maxima.nonzero().t()[2:4]


def extract_multiscale( net, img, detector, scale_f=2**0.25, 
                        min_scale=0.0, max_scale=1, 
                        min_size=256, max_size=1024, 
                        verbose=False):
    old_bm = torch.backends.cudnn.benchmark 
    torch.backends.cudnn.benchmark = False # speedup
    
    # extract keypoints at multiple scales
    B, three, H, W = img.shape
    assert B == 1 and three == 3, "should be a batch with a single RGB image"
    
    assert max_scale <= 1
    s = 1.0 # current scale factor
    
    X,Y,S,C,Q,D = [],[],[],[],[],[]
    while  s+0.001 >= max(min_scale, min_size / max(H,W)):
        if s-0.001 <= min(max_scale, max_size / max(H,W)):
            nh, nw = img.shape[2:]
            if verbose: print(f"extracting at scale x{s:.02f} = {nw:4d}x{nh:3d}")
            # extract descriptors
            with torch.no_grad():
                res = net(imgs=[img])
                
            # get output and reliability map
            descriptors = res['descriptors'][0]
            reliability = res['reliability'][0]
            repeatability = res['repeatability'][0]

            # normalize the reliability for nms
            # extract maxima and descs
            y,x = detector(**res) # nms
            c = reliability[0,0,y,x]
            q = repeatability[0,0,y,x]
            d = descriptors[0,:,y,x].t()
            n = d.shape[0]

            # accumulate multiple scales
            X.append(x.float() * W/nw)
            Y.append(y.float() * H/nh)
            S.append((32/s) * torch.ones(n, dtype=torch.float32, device=d.device))
            C.append(c)
            Q.append(q)
            D.append(d)
        s /= scale_f

        # down-scale the image for next iteration
        nh, nw = round(H*s), round(W*s)
        img = F.interpolate(img, (nh,nw), mode='bilinear', align_corners=False)

    # restore value
    torch.backends.cudnn.benchmark = old_bm

    Y = torch.cat(Y)
    X = torch.cat(X)
    S = torch.cat(S) # scale
    scores = torch.cat(C) * torch.cat(Q) # scores = reliability * repeatability
    XYS = torch.stack([X,Y,S], dim=-1)
    D = torch.cat(D)
    return XYS, D, scores


def extract_keypoints(args):
    iscuda = common.torch_set_gpu(args.gpu)

    # load the network...
    net = load_network(args.model)
    if iscuda: net = net.cuda()

    # create the non-maxima detector
    detector = NonMaxSuppression(
        rel_thr = args.reliability_thr, 
        rep_thr = args.repeatability_thr)

    while args.images:
        img_path = args.images.pop(0)
        
        if img_path.endswith('.txt'):
            args.images = open(img_path).read().splitlines() + args.images
            continue
        
        print(f"\nExtracting features for {img_path}")
        img = Image.open(img_path).convert('RGB')
        W, H = img.size
        img = norm_RGB(img)[None] 
        if iscuda: img = img.cuda()
        
        # extract keypoints/descriptors for a single image
        xys, desc, scores = extract_multiscale(net, img, detector,
            scale_f   = args.scale_f, 
            min_scale = args.min_scale, 
            max_scale = args.max_scale,
            min_size  = args.min_size, 
            max_size  = args.max_size, 
            verbose = True)

        xys = xys.cpu().numpy()
        desc = desc.cpu().numpy()
        scores = scores.cpu().numpy()
        idxs = scores.argsort()[-args.top_k or None:]
        
        outpath = img_path + '.' + args.tag
        print(f"Saving {len(idxs)} keypoints to {outpath}")
        np.savez(open(outpath,'wb'), 
            imsize = (W,H),
            keypoints = xys[idxs], 
            descriptors = desc[idxs], 
            scores = scores[idxs])



if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser("Extract keypoints for a given image")
    parser.add_argument("--model", type=str, required=True, help='model path')
    
    parser.add_argument("--images", type=str, required=True, nargs='+', help='images / list')
    parser.add_argument("--tag", type=str, default='r2d2', help='output file tag')
    
    parser.add_argument("--top-k", type=int, default=5000, help='number of keypoints')

    parser.add_argument("--scale-f", type=float, default=2**0.25)
    parser.add_argument("--min-size", type=int, default=256)
    parser.add_argument("--max-size", type=int, default=1024)
    parser.add_argument("--min-scale", type=float, default=0)
    parser.add_argument("--max-scale", type=float, default=1)
    
    parser.add_argument("--reliability-thr", type=float, default=0.7)
    parser.add_argument("--repeatability-thr", type=float, default=0.7)

    parser.add_argument("--gpu", type=int, nargs='+', default=[0], help='use -1 for CPU')
    args = parser.parse_args()

    extract_keypoints(args)




from e2cnn.nn import *
from e2cnn.group import *

import torch
import torch.nn as nn
import numpy as np


class E2SFCNN(torch.nn.Module):
    
    def __init__(self,
                 n_channels: int,
                 n_classes: int,
                 N: int = 16,
                 restrict: int = -1,
                 ):
        r"""
        
        Args:
            n_channels: number of channels in the input
            n_classes: number of output classes
            N: number of rotations of the equivariance group
            restrict: number of initial convolutional layers which are also flip equivariant.
                      After these layers, we restrict to only rotation equivariance.
                      By default (-1) the restriciton is never done so the model is flip and rotations equivariant.
                      If set to 0 the model is only rotation equivariant from the beginning
        """
        
        super(E2SFCNN, self).__init__()
        
        assert N > 1
        
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.N = N
        self.restrict = restrict
        
        gc = FlipRot2dOnR2(N)
        
        self.gc = gc
        
        self.LAYER = 0

        if self.restrict == self.LAYER:
            gc, _, _ = self.gc.restrict((None, N))
        
        r1 = FieldType(gc, [gc.trivial_repr] * n_channels)
        
        eq_layers = []
        
        # 28 px
        # Convolutional Layer 1

        self.LAYER += 1
        eq_layers += self.build_layers(r1, 24, 9, 0, None)
        
        # Convolutional Layer 2
        self.LAYER += 1
        eq_layers += self.build_layers(eq_layers[-1].out_type, 32, 7, 3, 2)
        
        # 14 px
        # Convolutional Layer 3
        self.LAYER += 1
        eq_layers += self.build_layers(eq_layers[-1].out_type, 36, 7, 3, None)

        # Convolutional Layer 4
        self.LAYER += 1
        eq_layers += self.build_layers(eq_layers[-1].out_type, 36, 7, 3, 2)

        # 7 px

        # Convolutional Layer 5
        self.LAYER += 1
        eq_layers += self.build_layers(eq_layers[-1].out_type, 64, 7, 3)

        # Convolutional Layer 6
        self.LAYER += 1
        eq_layers += self.build_layers(eq_layers[-1].out_type, 96, 5, 0, None, True)

        # Adaptive Pooling
        mpl = PointwiseAdaptiveMaxPool(eq_layers[-1].out_type, 1)
        eq_layers.append(mpl)

        # 1 px
        
        # c = 96
        c = eq_layers[-1].out_type.size
        
        self.in_repr = eq_layers[0].in_type
        self.eq_layers = SequentialModule(*eq_layers)

        # Fully Connected

        self.fully_net = nn.Sequential(
            nn.Dropout(p=0.3),
            nn.Linear(c, 96),
            nn.BatchNorm1d(96),
            nn.ELU(inplace=True),
    
            nn.Dropout(p=0.3),
            nn.Linear(96, 96),
            nn.BatchNorm1d(96),
            nn.ELU(inplace=True),
            
            nn.Dropout(p=0.3),
            nn.Linear(96, n_classes),
        )
    
    def forward(self, input):
        x = GeometricTensor(input, self.in_repr)
        
        features = self.eq_layers(x)

        features = features.tensor.reshape(x.tensor.shape[0], -1)
        
        out = self.fully_net(features)
        
        return out
        
    def build_layers(self,
                     r1: FieldType,
                     C: int,
                     s: int,
                     padding: int = 0,
                     pooling: int = None,
                     orientation_pooling: bool = False,
        ):

        gc = r1.isometries

        layers = []

        r2 = FieldType(gc, [gc.representations['regular']] * C)

        cl = R2Conv(r1,
                    r2,
                    s,
                    padding=padding
        )
        layers.append(cl)

        if self.restrict == self.LAYER:
            layers.append(RestrictionModule(layers[-1].out_type, (None, self.N)))
            layers.append(DisentangleModule(layers[-1].out_type))

        bn = InnerBatchNorm(layers[-1].out_type)
        layers.append(bn)

        if orientation_pooling:
            pl = CapsulePool(layers[-1].out_type)
            layers.append(pl)

        if pooling is not None:
            pl = PointwiseMaxPool(layers[-1].out_type, pooling)
            layers.append(pl)

        nnl = ReLU(layers[-1].out_type, inplace=True)
        layers.append(nnl)

        return layers




import warnings
from typing import TypedDict

import torch

from easy_local_features.submodules import git_dad as dad
from easy_local_features.submodules.git_dad.utils import get_best_device

from easy_local_features.matching.nearest_neighbor import NearestNeighborMatcher
from easy_local_features.feature.basemodel import BaseExtractor, MethodType
from easy_local_features.utils.ops import prepareImage

class DadConfig(TypedDict):
    num_keypoints: int
    resize: int
    nms_size: int


class DAD_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE
    default_conf = DadConfig(
        num_keypoints=1024,
        resize=1024,
        nms_size=3,
    )

    def __init__(self, conf={}):
        self.num_keypoints = conf.get("num_keypoints", self.default_conf["num_keypoints"])
        self.detector = dad.load_DaD()
        self.DEV = get_best_device()
        self.matcher = NearestNeighborMatcher()

    def detect(self, image, return_dict=None):
        img = prepareImage(image, imagenet=True, batch=True)
        img = image.to(self.DEV)
        

        mkpts = self.detector.detect({"image": img}, num_keypoints=self.num_keypoints)["keypoints"]
        mkpts = self.detector.to_pixel_coords(mkpts, img.shape[-2], img.shape[-1])

        if return_dict:
            return {"mkpts": mkpts}
        return mkpts

    def detectAndCompute(self, image, return_dict=None):
        raise NotImplementedError

    def compute(self, image, keypoints):
        raise NotImplementedError

    @property
    def has_detector(self):
        return True

    def to(self, device):
        self.detector.to(device)
        self.DEV = device


if __name__ == "__main__":
    from easy_local_features.utils import io, vis

    detector = DAD_baseline({"num_keypoints": 512})

    img0 = io.fromPath("tests/assets/megadepth0.jpg")
    img1 = io.fromPath("tests/assets/megadepth1.jpg")

    kps0 = detector.detect(img0)
    kps1 = detector.detect(img1)

    vis.plot_pair(img0, img1)
    vis.plot_keypoints(keypoints0=kps0.cpu(), keypoints1=kps1.cpu())
    vis.show()



import torch
import torch.nn as nn

from .utils import cut_to_match, size_is_pow2
from .ops import NoOp

class Conv(nn.Sequential):
    def __init__(self, in_, out_, size, setup=None):
        norm = setup['norm'](in_)
        nonl = setup['gate'](in_)
        dropout = setup['dropout']()

        if setup['padding']:
            padding = size // 2
        else:
            padding = 0

        if 'bias' in setup:
            bias = setup['bias']
        else:
            bias = True

        conv = nn.Conv2d(in_, out_, size, padding=padding, bias=bias)

        super(Conv, self).__init__(norm, nonl, dropout, conv)


class Upsample(nn.Sequential):
    def __init__(self, n_features, size, setup=None):
        conv_kwargs = {
            'stride': 2,
            'output_padding': 1
        }
        norm = setup['norm'](n_features)
        nonl = setup['gate'](n_features)
        conv = nn.ConvTranspose2d(
            n_features, n_features, size, conv_kwargs=conv_kwargs
        )

        super(Upsample, self).__init__(norm, nonl, conv)


class Downsample(nn.Sequential):
    def __init__(self, n_features, size, setup=None):

        conv_kwargs = {
            'stride': 2,
            'padding': size // 2
        }
        norm = setup['norm'](n_features)
        nonl = setup['gate'](n_features)
        conv = nn.Conv2d(
            n_features, n_features, size, conv_kwargs=conv_kwargs
        )

        super(Downsample, self).__init__(norm, nonl, conv)


class UnetDownBlock(nn.Sequential):
    def __init__(self, in_, out_, size=5, name=None, is_first=False,
                 setup=None):

        self.name = name
        self.in_ = in_
        self.out_ = out_
        
        if is_first:
            downsample = NoOp()
            conv1 = Conv(in_, out_, size, setup={**setup, 'gate': NoOp, 'norm': NoOp})
        else:
            downsample = setup['downsample'](in_, size, setup=setup)
            conv1 = Conv(in_, out_, size, setup=setup)

        conv2 = Conv(out_, out_, size, setup=setup)
        super(UnetDownBlock, self).__init__(downsample, conv1, conv2)


class ThinUnetDownBlock(nn.Sequential):
    def __init__(self, in_, out_, size=5, name=None, is_first=False,
                 setup=None):

        self.name = name
        self.in_ = in_
        self.out_ = out_
        
        if is_first:
            downsample = NoOp()
            conv = Conv(in_, out_, size, setup={**setup, 'gate': NoOp, 'norm': NoOp})
        else:
            downsample = setup['downsample'](in_, size, setup=setup)
            conv = Conv(in_, out_, size, setup=setup)

        super(ThinUnetDownBlock, self).__init__(downsample, conv)


class UnetUpBlock(nn.Module):
    def __init__(self, bottom_, horizontal_, out_,
                 size=5, name=None, setup=None):

        super(UnetUpBlock, self).__init__()

        self.name = name
        self.bottom_ = bottom_
        self.horizontal_ = horizontal_
        self.cat_ = bottom_ + horizontal_
        self.out_ = out_

        self.upsample = setup['upsample'](bottom_, size, setup=setup)

        conv1 = Conv(self.cat_, self.cat_, size, setup=setup)
        conv2 = Conv(self.cat_, self.out_, size, setup=setup,)
        self.seq = nn.Sequential(conv1, conv2)


    def forward(self, bot,hor):

        bot_big = self.upsample(bot)
        hor = cut_to_match(bot_big, hor, n_pref=2)
        combined = torch.cat([bot_big, hor], dim=1)

        return self.seq(combined)


class ThinUnetUpBlock(nn.Module):
    def __init__(self, bottom_, horizontal_, out_,
                 size=5, name=None, setup=None):

        super(ThinUnetUpBlock, self).__init__()

        self.name = name
        self.bottom_ = bottom_
        self.horizontal_ = horizontal_
        self.cat_ = bottom_ + horizontal_
        self.out_ = out_

        self.upsample = setup['upsample'](bottom_, size, setup=setup)
        self.conv = Conv(self.cat_, self.out_, size, setup=setup)


    def forward(self, bot, hor):


        bot_big = self.upsample(bot)
        hor = cut_to_match(bot_big, hor, n_pref=2)
        combined = torch.cat([bot_big, hor], dim=1)

        return self.conv(combined)



"""Pytest configuration to ensure the package in ``src/`` is importable.

Now that tests live in a top-level ``tests/`` folder (outside ``src``), we need
to ensure ``src`` is on ``sys.path`` when invoking ``pytest`` without an
editable install.
"""

from __future__ import annotations
import sys
from pathlib import Path

REPO_ROOT = Path(__file__).resolve().parent.parent
SRC_DIR = REPO_ROOT / "src"
if SRC_DIR.is_dir():
    sys.path.insert(0, str(SRC_DIR))
